"""
Use dictionaries to create train and test domain split
"""
from domainlab.tasks.b_task_classif import NodeTaskDictClassif  # abstract class


def mk_task_dset(
    isize,
    taskna="task_custom",  # name of the task
    dim_y=None,
    list_str_y=None,
    parent=NodeTaskDictClassif,
    succ=None,
):
    """
    make a task via a dictionary of dataset where the key is domain
    value is a tuple of dataset for training and dataset for
    validation (can be identical to training)
    """

    class NodeTaskDset(parent):
        """
        Use dictionaries to create train and test domain split
        """

        def conf_without_args(self):
            """
            set member variables
            """
            self._name = taskna

            if list_str_y is None and dim_y is None:
                raise RuntimeError(
                    "arguments list_str_y and dim_y can not be both None!"
                )

            self.list_str_y = (
                list_str_y  # list_str_y has to be initialized before dim_y
            )
            self.dim_y = dim_y

            if self.list_str_y is None:
                self.list_str_y = [f"class{ele}" for ele in range(0, self.dim_y)]
            self.isize = isize

        def get_dset_by_domain(self, args, na_domain, split=False):
            """
            each domain correspond to one dataset, must be implemented by child class
            """
            return self.dict_dset_all[na_domain]

        def init_business(self, args, trainer=None):
            """
            create a dictionary of datasets
            """
            self.set_list_domains(list(self.dict_dset_all.keys()))
            super().init_business(args, trainer)

        def add_domain(self, name, dset_tr, dset_val=None):
            """
            add domain via, name, dataset and transformations
            """
            self.dict_dset_all[name] = (dset_tr, dset_val)
            # when add a new domain, change self state
            self.set_list_domains(list(self.dict_dset_all.keys()))

    return NodeTaskDset(succ=succ)



"""
Builder
1. classifier for domain **and** class
2. p(z_y|y) **and** p(z_d|d)
"""
from domainlab.compos.nn_zoo.net_classif import ClassifDropoutReluLinear
from domainlab.compos.vae.a_vae_builder import AbstractVAEBuilderChainNode
from domainlab.compos.vae.compos.decoder_cond_prior import (
    LSCondPriorLinearBnReluLinearSoftPlus,
)


class ChainNodeVAEBuilderClassifCondPrior(AbstractVAEBuilderChainNode):
    """
    1. This class defines common methods shared by child classes:
        - classifier for domain/class
        - conditional prior
    2. Bridge pattern: separate abstraction (vae model) and implementation)
    """

    def construct_classifier(self, input_dim, output_dim):
        """
        classifier can be used to both classify class-label and domain-label
        @param input_dim: can be both zy_dim or zd_dim
        """
        return ClassifDropoutReluLinear(input_dim, output_dim)

    def construct_cond_prior(self, input_dim, output_dim):
        """
        For both p(z_y|y) and p(z_d|d)
        """
        net_p_z_pars = LSCondPriorLinearBnReluLinearSoftPlus(input_dim, output_dim)
        return net_p_z_pars

    def build_encoder(self):
        raise NotImplementedError

    def build_decoder(self):
        raise NotImplementedError

    def is_myjob(self, request):
        return NotImplementedError



"""
build hduva model, get trainer from cmd arguments
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.observers.c_obvisitor_cleanup import ObVisitorCleanUp
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.compos.pcr.request import RequestVAEBuilderCHW
from domainlab.compos.vae.utils_request_chain_builder import VAEChainNodeGetter
from domainlab.models.model_hduva import mk_hduva
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderHDUVA(NodeAlgoBuilder):
    """
    NodeAlgoBuilderHDUVA
    """

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        task = exp.task
        args = exp.args
        task.get_list_domains_tr_te(args.tr_d, args.te_d)
        request = RequestVAEBuilderCHW(task.isize.c, task.isize.h, task.isize.w, args)
        device = get_device(args)
        node = VAEChainNodeGetter(request, args.topic_dim)()
        model = mk_hduva()(
            node,
            zd_dim=args.zd_dim,
            zy_dim=args.zy_dim,
            zx_dim=args.zx_dim,
            device=device,
            topic_dim=args.topic_dim,
            list_str_y=task.list_str_y,
            gamma_d=args.gamma_d,
            gamma_y=args.gamma_y,
            beta_t=args.beta_t,
            beta_x=args.beta_x,
            beta_y=args.beta_y,
            beta_d=args.beta_d,
        )
        model = self.init_next_model(model, exp)
        model_sel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
        observer = ObVisitorCleanUp(ObVisitor(model_sel))
        trainer = TrainerChainNodeGetter(args.trainer)(default="hyperscheduler")
        trainer.init_business(model, task, observer, device, args)
        return trainer, model, observer, device



"""
gridsearch for the hyperparameter space

def add_next_param_from_list is an recursive function to make cartesian product along all the scalar hyper-parameters, this resursive function is used
in def grid_task

"""
import copy
import json
import os
import warnings

import numpy as np
import pandas as pd

import domainlab.utils.hyperparameter_sampling as sampling
from domainlab.utils.get_git_tag import get_git_tag
from domainlab.utils.logger import Logger

G_MODEL_NA = "model"
G_METHOD_NA = "method"


def round_to_discreate_grid_uniform(grid, param_config):
    """
    round the values of the grid to the grid spacing specified in the config
    for uniform and loguniform grids
    """
    if float(param_config["step"]) == 0:
        return grid
    mini = float(param_config["min"])
    maxi = float(param_config["max"])
    if maxi - mini < float(param_config["step"]):
        raise RuntimeError(
            "distance between max and min to small for defined step size"
        )

    discreate_gird = np.arange(
        mini, maxi + float(param_config["step"]), step=float(param_config["step"])
    )
    for num, elem in enumerate(list(grid)):
        # search for the closest allowed grid point to the scalar elem
        grid[num] = discreate_gird[(np.abs(discreate_gird - elem)).argmin()]
    grid_unique = np.unique(grid)
    grid_out = grid_unique
    return grid_out


def round_to_discreate_grid_normal(grid, param_config):
    """
    round the values of the grid to the grid spacing specified in the config
    for normal and lognormal grids
    """
    if float(param_config["step"]) == 0:
        return grid
    # for normal and lognormal no min and max is provided
    # in this case the grid is constructed around the mean
    neg_steps = np.ceil(
        (float(param_config["mean"]) - np.min(grid)) / float(param_config["step"])
    )
    pos_steps = np.ceil(
        (np.max(grid) - float(param_config["mean"])) / float(param_config["step"])
    )
    mini = float(param_config["mean"]) - float(param_config["step"]) * neg_steps
    maxi = float(param_config["mean"]) + float(param_config["step"]) * pos_steps

    discreate_gird = np.arange(mini, maxi, step=float(param_config["step"]))
    for num, elem in enumerate(list(grid)):
        grid[num] = discreate_gird[(np.abs(discreate_gird - elem)).argmin()]
    return np.unique(grid)


def uniform_grid(param_config):
    """
    get a uniform distributed grid given the specifications in the param_config
    param_config: config which needs to contain 'num', 'max', 'min', 'step'
    """
    num = int(param_config["num"])
    maxi = float(param_config["max"])
    mini = float(param_config["min"])
    step = (maxi - mini) / num
    # linspace does include the end of the interval and include the beginning
    # we move away from mini and maxi to sample inside the open interval (mini, maxi)
    grid = np.linspace(mini + step / 2, maxi - step / 2, num)
    if "step" in param_config.keys():
        return round_to_discreate_grid_uniform(grid, param_config)
    return grid


def loguniform_grid(param_config):
    """
    get a loguniform distributed grid given the specifications in the param_config
    param_config: config which needs to contain 'num', 'max', 'min'
    """
    num = int(param_config["num"])
    maxi = np.log10(float(param_config["max"]))
    mini = np.log10(float(param_config["min"]))
    step = (maxi - mini) / num
    # linspace does exclude the end of the interval and include the beginning
    grid = 10 ** np.linspace(mini + step / 2, maxi - step / 2, num)
    if "step" in param_config.keys():
        return round_to_discreate_grid_uniform(grid, param_config)
    return grid


def normal_grid(param_config, lognormal=False):
    """
    get a normal distributed grid given the specifications in the param_config
    param_config: config which needs to contain 'num', 'mean', 'std'
    """
    if int(param_config["num"]) == 1:
        return np.array([float(param_config["mean"])])
    # Box–Muller transform to get from a uniform distribution to a normal distribution
    num = int(np.floor(int(param_config["num"]) / 2))
    step = 2 / (int(param_config["num"]) + 1)
    # for a even number of samples
    if int(param_config["num"]) % 2 == 0:
        param_grid = np.arange(step, 1, step=step)[:num]
        stnormal_grid = np.sqrt(-2 * np.log(param_grid))
        stnormal_grid = np.append(stnormal_grid, -stnormal_grid)
        stnormal_grid = stnormal_grid / np.std(stnormal_grid)
        stnormal_grid = float(param_config["std"]) * stnormal_grid + float(
            param_config["mean"]
        )
    # for a odd number of samples
    else:
        param_grid = np.arange(step, 1, step=step)[:num]
        stnormal_grid = np.sqrt(-2 * np.log(param_grid))
        stnormal_grid = np.append(stnormal_grid, -stnormal_grid)
        stnormal_grid = np.append(stnormal_grid, 0)
        stnormal_grid = stnormal_grid / np.std(stnormal_grid)
        stnormal_grid = float(param_config["std"]) * stnormal_grid + float(
            param_config["mean"]
        )

    if "step" in param_config.keys() and lognormal is False:
        return round_to_discreate_grid_normal(stnormal_grid, param_config)
    return stnormal_grid


def lognormal_grid(param_config):
    """
    get a normal distributed grid given the specifications in the param_config
    param_config: config which needs to contain 'num', 'mean', 'std'
    """
    grid = 10 ** normal_grid(param_config, lognormal=True)
    if "step" in param_config.keys():
        return round_to_discreate_grid_normal(grid, param_config)
    return grid


def add_next_param_from_list(param_grid: dict, grid: dict, grid_df: pd.DataFrame):
    """
    can be used in a recoursive fassion to add all combinations of the parameters in
    param_grid to grid_df
    param_grid: dictionary with all possible values for each parameter
                {'p1': [1, 2, 3], 'p2': [0, 5], ...}
    grid: a grid which will build itself in the recursion, start with grid = {}
            after one step grid = {p1: 1}
    grid_df: dataframe which will save the finished grids
    task_name: task name
    also: G_MODEL_NA name
    """
    if len(param_grid.keys()) != 0:
        # specify the parameter to be used
        param_name = list(param_grid.keys())[0]
        # for all values of this parameter perform
        for param in param_grid[param_name]:
            # add the parameter to the grid
            grid_new = dict(grid)
            grid_new.update({param_name: param})
            # remove the parameter from param_grid
            param_grid_new = dict(param_grid)
            param_grid_new.pop(param_name)
            # resume with the next parameter
            add_next_param_from_list(param_grid_new, grid_new, grid_df)
    else:
        # add sample to grid_df
        grid_df.loc[len(grid_df.index)] = [grid]


def add_references_and_check_constraints(
    grid_df_prior, grid_df, referenced_params, config, task_name
):
    """
    in the last step all parameters which are referenced need to be add to the
    grid. All gridpoints not satisfying the constraints are removed afterwards.
    """
    for dictio in grid_df_prior["params"]:
        for key, val in dictio.items():
            exec(f"{key} = val")
        # add referenced params
        for rev_param, val in referenced_params.items():
            val = eval(val)
            dictio.update({rev_param: val})
            exec(f"{rev_param} = val")
        # check constraints
        if "hyperparameters" in config.keys():
            constraints = config["hyperparameters"].get("constraints", None)
        else:
            constraints = config.get("constraints", None)
        if constraints is not None:
            accepted = True
            for constr in constraints:
                if not eval(constr):
                    accepted = False
            if accepted:
                grid_df.loc[len(grid_df.index)] = [task_name, config["model"], dictio]
        else:
            grid_df.loc[len(grid_df.index)] = [task_name, config["model"], dictio]


def sample_grid(param_config):
    """
    given the parameter config, this function samples all parameters which are distributed
    according the the categorical, uniform, loguniform, normal or lognormal distribution.
    """
    # sample cathegorical parameter
    if param_config["distribution"] == "categorical":
        param_grid = sampling.CategoricalHyperparameter("", param_config).allowed_values
    # sample uniform parameter
    elif param_config["distribution"] == "uniform":
        param_grid = uniform_grid(param_config)
    # sample loguniform parameter
    elif param_config["distribution"] == "loguniform":
        param_grid = loguniform_grid(param_config)
    # sample normal parameter
    elif param_config["distribution"] == "normal":
        param_grid = normal_grid(param_config)
    # sample lognormal parameter
    elif param_config["distribution"] == "lognormal":
        param_grid = lognormal_grid(param_config)
    else:
        raise RuntimeError(
            f'distribution "{param_config["distribution"]}" not '
            f"implemented use a distribution from "
            f"[categorical, uniform, loguniform, normal, lognormal]"
        )

    # ensure that the gird does have the correct datatype
    # (only check for int, othervise float is used)
    if "datatype" in param_config.keys():
        if param_config["datatype"] == "int":
            param_grid = np.array(param_grid)
            param_grid = param_grid.astype(int)
        # NOTE: converting int to float will cause error for VAE, avoid do
        # it here
        return param_grid


def build_param_grid_of_shared_params(shared_df):
    """
    go back from the data frame format of the shared hyperparamters to a list format
    """
    if shared_df is None:
        return None
    shared_grid = {}
    for key in shared_df["params"].iloc[0].keys():
        grid_points = []
        for i in shared_df["params"].keys():
            grid_points.append(shared_df["params"][i][key])
        shared_grid[key] = np.array(grid_points)
    return shared_grid


def rais_error_if_num_not_specified(param_name: str, param_config: dict):
    """
    for each parameter a number of grid points needs to be specified
    This function raises an error if this is not the case
    param_name: parameter name under consideration
    param_config: config of this parameter
    """
    if not param_name == "constraints":
        if (
            not "num" in param_config.keys()
            and not "reference" in param_config.keys()
            and not param_config["distribution"] == "categorical"
        ):
            raise RuntimeError(
                f"the number of parameters in the grid direction "
                f"of {param_name} needs to be specified"
            )


def add_shared_params_to_param_grids(shared_df, dict_param_grids, config):
    """
    use the parameters in the dataframe of shared parameters and add them
    to the dictionary of parameters for the current task
    only the shared parameters specified in the config are respected
    shared_df: Dataframe of shared hyperparameters
    dict_param_grids: dictionary of the parameter grids
    config: config for the current task
    """
    dict_shared_grid = build_param_grid_of_shared_params(shared_df)
    if "shared" in config.keys():
        list_names = config["shared"]
        dict_shared_grid = {key: dict_shared_grid[key] for key in config["shared"]}
        if dict_shared_grid is not None:
            for key in dict_shared_grid.keys():
                dict_param_grids[key] = dict_shared_grid[key]
    return dict_param_grids


def grid_task(
    grid_df: pd.DataFrame, task_name: str, config: dict, shared_df: pd.DataFrame
):
    """create grid for one sampling task for a method and add it to the dataframe"""
    if "hyperparameters" in config.keys():
        dict_param_grids = {}
        referenced_params = {}
        for param_name in config["hyperparameters"].keys():
            param_config = config["hyperparameters"][param_name]
            rais_error_if_num_not_specified(param_name, param_config)

            # constraints are not parameters
            if not param_name == "constraints":
                # remember all parameters which are reverenced
                if "datatype" not in param_config.keys():
                    warnings.warn(
                        f"datatype not specified in {param_config} \
                                  for {param_name}, take float as default"
                    )
                    param_config["datatype"] = "float"

                if "reference" in param_config.keys():
                    referenced_params.update({param_name: param_config["reference"]})
                # sample other parameter
                elif param_name != "constraints":
                    dict_param_grids.update({param_name: sample_grid(param_config)})

        # create the grid from the individual parameter grids
        # constraints are not respected in this step
        grid_df_prior = pd.DataFrame(columns=["params"])
        # add shared parameters to dict_param_grids
        dict_param_grids = add_shared_params_to_param_grids(
            shared_df, dict_param_grids, config
        )
        add_next_param_from_list(dict_param_grids, {}, grid_df_prior)

        # add referenced params and check constraints
        add_references_and_check_constraints(
            grid_df_prior, grid_df, referenced_params, config, task_name
        )
        if grid_df[grid_df[G_MODEL_NA] == config["model"]].shape[0] == 0:
            raise RuntimeError(
                "No valid value found for this grid spacing, refine grid"
            )
        return grid_df
    elif "shared" in config.keys():
        shared_grid = shared_df.copy()
        shared_grid[G_MODEL_NA] = config["model"]
        shared_grid[G_METHOD_NA] = task_name
        if "constraints" in config.keys():
            config["hyperparameters"] = {"constraints": config["constraints"]}
        add_references_and_check_constraints(
            shared_grid, grid_df, {}, config, task_name
        )
        return grid_df
    else:
        # add single line if no varying hyperparameters are specified.
        grid_df.loc[len(grid_df.index)] = [task_name, config["model"], {}]
        return grid_df


def sample_gridsearch(config: dict, dest: str = None) -> pd.DataFrame:
    """
    create the hyperparameters grid according to the given
    config, which should be the dictionary of the full
    benchmark config yaml.
    Result is saved to 'output_dir/hyperparameters.csv' of the
    config if not specified explicitly.

    Note: Parts of the yaml content are executed. Thus use this
    only with trusted config files.
    """
    if dest is None:
        dest = config["output_dir"] + os.sep + "hyperparameters.csv"

    logger = Logger.get_logger()
    samples = pd.DataFrame(columns=[G_METHOD_NA, G_MODEL_NA, "params"])
    shared_samples_full = pd.DataFrame(columns=[G_METHOD_NA, G_MODEL_NA, "params"])

    if "Shared params" in config.keys():
        shared_val = {"model": "all", "hyperparameters": config["Shared params"]}
        # fill up the dataframe shared samples
        shared_samples_full = grid_task(shared_samples_full, "all", shared_val, None)
    else:
        shared_samples_full = None
    for key, val in config.items():
        if sampling.is_dict_with_key(val, "model"):
            if shared_samples_full is not None:
                shared_samples = shared_samples_full.copy(deep=True)
                if "shared" in val.keys():
                    shared = val["shared"]
                else:
                    shared = []
                for line_num in range(shared_samples.shape[0]):
                    hyper_p_dict = shared_samples.iloc[line_num]["params"].copy()
                    key_list = copy.deepcopy(list(hyper_p_dict.keys()))
                    if not all(x in key_list for x in shared):
                        raise RuntimeError(
                            f"shared keys: {shared} not included in global shared keys {key_list}"
                        )
                    for key_ in key_list:
                        if key_ not in shared:
                            del hyper_p_dict[key_]
                    shared_samples.iloc[line_num]["params"] = hyper_p_dict
                # remove all duplicates
                shared_samples = shared_samples.drop_duplicates(subset="params")
            else:
                shared_samples = None

            samples = grid_task(samples, key, val, shared_samples)
            logger.info(
                f"number of gridpoints for {key} : "
                f'{samples[samples[G_MODEL_NA] == val["model"]].shape[0]}'
            )

    os.makedirs(os.path.dirname(dest), exist_ok=True)
    logger.info(f"number of total sampled gridpoints: {samples.shape[0]}")
    samples.to_csv(dest)
    # create a txt file with the commit information
    with open(
        config["output_dir"] + os.sep + "commit.txt", "w", encoding="utf8"
    ) as file:
        file.writelines("use git log |grep \n")
        file.writelines("consider remove leading b in the line below \n")
        file.write(get_git_tag())
    with open(
        config["output_dir"] + os.sep + "config.txt", "w", encoding="utf8"
    ) as file:
        json.dump(config, file)
    return samples



"""
Abstract class for TaskClassif
"""
import os

from domainlab.tasks.a_task import NodeTaskDG
from domainlab.tasks.utils_task import img_loader2dir


class NodeTaskDGClassif(NodeTaskDG):
    """
    abstract class for classification task
    """

    def __init__(self, succ=None):
        # just for declaration of variables
        self._list_str_y = None
        self._dim_y = None
        # super() must come last instead of in the beginning
        super().__init__(succ)

    @property
    def list_str_y(self):
        """
        getter for list_str_y
        """
        return self._list_str_y

    @list_str_y.setter
    def list_str_y(self, list_str_y):
        """
        setter for list_str_y
        """
        self._list_str_y = list_str_y

    @property
    def dim_y(self):
        """classification dimension"""
        if self._dim_y is None:
            if self.list_str_y is not None:
                return len(self.list_str_y)
            raise RuntimeError("attribute list_str_y not set yet")
        return self._dim_y

    @dim_y.setter
    def dim_y(self, dim_y):
        """
        setter for dim_y, for classification task, the number of labels to predict
        """
        if self.list_str_y is not None:
            if len(self.list_str_y) is not dim_y:
                raise RuntimeError(
                    f"dim y={dim_y} not equal to self.list_str_y={self.list_str_y}"
                )
        self._dim_y = dim_y

    def sample_sav(self, root, batches=5, subfolder_na="task_sample"):
        """
        sample data from task and save to disk
        """
        folder_na = os.path.join(root, self.task_name, subfolder_na)

        img_loader2dir(
            self.loader_te,
            list_domain_na=self.get_list_domains(),
            list_class_na=self.list_str_y,
            folder=folder_na,
            batches=batches,
            test=True,
        )

        img_loader2dir(
            self.loader_tr,
            list_domain_na=self.get_list_domains(),
            list_class_na=self.list_str_y,
            folder=folder_na,
            batches=batches,
        )



"""
experiment
"""
import datetime
import os
import warnings

from domainlab.algos.zoo_algos import AlgoBuilderChainNodeGetter
from domainlab.exp.exp_utils import AggWriter
from domainlab.tasks.zoo_tasks import TaskChainNodeGetter
from domainlab.utils.logger import Logger
from domainlab.utils.sanity_check import SanityCheck

os.environ["CUDA_LAUNCH_BLOCKING"] = "1"  # debug


class Exp:
    """
    Exp is combination of Task, Algorithm, and Configuration (including random seed)
    """

    def __init__(self, args, task=None, model=None, observer=None, visitor=AggWriter):
        """
        :param args:
        :param task: default None
        :param model: default None
        """
        self.task = task
        self.curr_dir = os.getcwd()
        if task is None:
            self.task = TaskChainNodeGetter(args)()

        self.args = args
        algo_builder = AlgoBuilderChainNodeGetter(
            self.args.model, self.args.apath
        )()  # request
        # the critical logic below is to avoid circular dependence between task initialization
        # and trainer initialization:

        self.trainer, self.model, observer_default, device = algo_builder.init_business(
            self
        )
        # sanity check has to be done after init_business
        # jigen algorithm builder has method dset_decoration_args_algo, which could AOP
        # into the task intilization process
        if args.san_check:
            sancheck = SanityCheck(args, self.task)
            sancheck.dataset_sanity_check()

        if model is not None:
            self.model = model
        self.epochs = self.args.epos
        self.epoch_counter = 1
        if observer is None:
            observer = observer_default
        if not self.trainer.flag_initialized:
            self.trainer.init_business(self.model, self.task, observer, device, args)
        self.visitor = visitor(self)  # visitor depends on task initialization first
        # visitor must be initialized last after trainer is initialized
        self.model.set_saver(self.visitor)

    def execute(self, num_epochs=None):
        """
        train model
        check performance by loading persisted model
        """
        if num_epochs is None:
            num_epochs = self.epochs + 1
        t_0 = datetime.datetime.now()
        logger = Logger.get_logger()
        logger.info(f"\n Experiment start at: {str(t_0)}")
        t_c = t_0
        self.trainer.before_tr()
        for epoch in range(1, num_epochs):
            t_before_epoch = t_c
            flag_stop = self.trainer.tr_epoch(epoch)
            t_c = datetime.datetime.now()
            logger.info(
                f"after epoch: {epoch},"
                f"now: {str(t_c)},"
                f"epoch time: {t_c - t_before_epoch},"
                f"used: {t_c - t_0},"
                f"model: {self.visitor.model_name}"
            )
            logger.info(f"working direcotry: {self.curr_dir}")
            # current time, time since experiment start, epoch time
            if flag_stop:
                self.epoch_counter = epoch
                logger.info("early stop trigger")
                break
            if epoch == self.epochs:
                self.epoch_counter = self.epochs
            else:
                self.epoch_counter += 1
        logger.info(
            f"Experiment finished at epoch: {self.epoch_counter} "
            f"with time: {t_c - t_0} at {t_c}"
        )
        self.trainer.post_tr()

    def clean_up(self):
        """
        to be called by a decorator
        """
        try:
            # oracle means use out-of-domain test accuracy to select the model
            self.visitor.remove("oracle")  # pylint: disable=E1101
        except FileNotFoundError:
            pass

        try:
            # the last epoch:
            # have a model to evaluate in case the training stops in between
            self.visitor.remove("epoch")  # pylint: disable=E1101
        except FileNotFoundError:
            logger = Logger.get_logger()
            logger.warn("failed to remove model_epoch: file not found")
            warnings.warn("failed to remove model_epoch: file not found")

        try:
            # without suffix: the selected model
            self.visitor.remove()  # pylint: disable=E1101
        except FileNotFoundError:
            logger = Logger.get_logger()
            logger.warn("failed to remove model")
            warnings.warn("failed to remove model")

        try:
            # for matchdg
            self.visitor.remove("ctr")  # pylint: disable=E1101
        except FileNotFoundError:
            pass



"""
The class TaskPathList provides the user an interface to provide a file with
each line consisting of a pair, where the first slot contains the path
(either absolute or relative if the user knows from where this package is
executed)
of an image and the second slot contains the class label as a numerical string.
"""
import os

import torch.multiprocessing

from domainlab.dsets.dset_img_path_list import DsetImPathList
from domainlab.dsets.utils_data import mk_fun_label2onehot
from domainlab.tasks.b_task_classif import NodeTaskDictClassif

torch.multiprocessing.set_sharing_strategy("file_system")
# "too many opened files" https://github.com/pytorch/pytorch/issues/11201


class NodeTaskPathListDummy(NodeTaskDictClassif):
    """
    typedef class so that other function can use isinstance
    """

    def get_dset_by_domain(self, args, na_domain, split=False):
        raise NotImplementedError


def mk_node_task_path_list(
    isize,
    img_trans_te,
    list_str_y,
    img_trans_tr,
    dict_class_label_ind2name,
    dict_domain2imgroot,
    dict_d2filepath_list_img_tr,
    dict_d2filepath_list_img_val,
    dict_d2filepath_list_img_te,
    succ=None,
):
    """mk_node_task_path_list.

    :param isize:
    :param list_str_y:
    :param img_trans_tr:
    :param dict_class_label_ind2name:
    :param dict_domain2imgroot:
    :param dict_d2filepath_list_img_tr:
    :param dict_d2filepath_list_img_val:
    :param dict_d2filepath_list_img_te:
    :param succ:
    """

    class NodeTaskPathList(NodeTaskPathListDummy):
        """
        The class TaskPathList provides the user an interface to provide a file
        with each line consisting of a pair separated by comma, where the
        first slot contains the path (either absolute or relative if the user
        knows from where this package is executed) of an image and the second
        slot contains the class label as a numerical string.
        e.g.: /path/2/file/art_painting/dog/pic_376.jpg 1
        """

        def _get_complete_domain(self, na_domain, dict_domain2pathfilepath):
            """_get_complete_domain.

            :param na_domain:
            """
            if na_domain not in self.list_domain_tr:
                trans = img_trans_te
            else:
                if self._dict_domain_img_trans:
                    trans = self._dict_domain_img_trans[na_domain]
                else:
                    trans = img_trans_tr
            root_img = self.dict_domain2imgroot[na_domain]
            path2filelist = dict_domain2pathfilepath[na_domain]
            path2filelist = os.path.expanduser(path2filelist)
            root_img = os.path.expanduser(root_img)
            dset = DsetImPathList(
                root_img,
                path2filelist,
                trans_img=trans,
                trans_target=mk_fun_label2onehot(len(self.list_str_y)),
            )
            return dset

        def get_dset_by_domain(self, args, na_domain, split=True):
            """get_dset_by_domain.

            :param args:
            :param na_domain:
            :param split: for test set, use the whole
            """
            if not split:  # no train/val split for test domain
                # the user is required to input tr, val, te file path
                # if split=False, then only te is used, which contains
                # the whole dataset
                dset = self._get_complete_domain(
                    na_domain, self._dict_domain2filepath_list_im_te
                )
                # test set contains train+validation
                return dset, dset  # @FIXME: avoid returning two identical

            dset = self._get_complete_domain(
                na_domain,
                # read training set from user configuration
                self._dict_domain2filepath_list_im_tr,
            )

            dset_val = self._get_complete_domain(
                na_domain,
                # read validation set from user configuration
                self._dict_domain2filepath_list_im_val,
            )

            return dset, dset_val

        def conf(self):
            """
            set task attribute in initialization
            """
            self.list_str_y = list_str_y
            self.isize = isize
            self.dict_class_label_ind2name = dict_class_label_ind2name
            self.dict_domain2imgroot = dict_domain2imgroot
            self._dict_domain2filepath_list_im_tr = dict_d2filepath_list_img_tr
            self._dict_domain2filepath_list_im_val = dict_d2filepath_list_img_val
            self._dict_domain2filepath_list_im_te = dict_d2filepath_list_img_te
            self.set_list_domains(list(self.dict_domain2imgroot.keys()))

        def __init__(self, succ=None):
            super().__init__(succ)
            self.conf()

    return NodeTaskPathList(succ)



import numpy as np
import torch
from torch.nn import functional as F


def mk_dummy_label_list_str(prefix, dim):
    """
    only used for testing, to generate list of class/domain label names
    """
    return [prefix + str(i) for i in range(dim)]


def logit2preds_vpic(logit):
    """
    :logit: batch of logit vector
    :return: vector of one-hot,
             vector of probability,
             index,
             maximum probability
    """
    mat_prob = F.softmax(logit, dim=1)
    # get the index of the maximum softmax probability
    max_prob, max_ind = torch.topk(mat_prob, 1)
    # convert the digit(s) to one-hot tensor(s)
    one_hot = logit.new_zeros(mat_prob.size())
    one_hot = one_hot.scatter_(dim=1, index=max_ind, value=1.0)
    return one_hot, mat_prob, max_ind, max_prob


def get_label_na(tensor_ind, list_str_na):
    """
    given list of label names in strings, map tensor of index to label names
    """
    arr_ind_np = tensor_ind.cpu().numpy()
    arr_ind = np.squeeze(
        arr_ind_np, axis=1
    )  # explicitly use axis=1 to deal with edge case of only
    # instance left
    # list_ind = list(arr_ind): if there is only dimension 1 tensor_ind, then there is a problem
    list_ind = arr_ind.tolist()
    list_na = [list_str_na[ind] for ind in list_ind]
    return list_na



"""
convert ids to a list of domain names
"""


def parse_domain_id(list_domain_id, list_domains):
    """
    Convert ids to a list of domain names.
    :param list_domain_id: domain id or ids provided as an int or str,
    or a list of int or str.
    :param list_domains: list of available domains
    :return: list of domain names
    """
    if not isinstance(list_domain_id, list):
        list_domain_id = [list_domain_id]
    list_domains_subset = []
    for ele in list_domain_id:
        if isinstance(ele, int):
            list_domains_subset.append(list_domains[ele])
        elif isinstance(ele, str):
            if ele.isdigit():
                list_domains_subset.append(list_domains[int(ele)])
            else:
                list_domains_subset.append(ele)
        else:
            raise RuntimeError("domain ids should be either int or str")
    return list_domains_subset



import torch.nn as nn


class LayerId(nn.Module):
    """
    used to delete layers
    """

    def __init__(self):
        super().__init__()

    def forward(self, x):
        """
        :param x:
        """
        return x


class DenseNet(nn.Module):
    """
    (input)-dropout-linear-relu-dropout-linear-relu(output)
    """

    def __init__(
        self, input_flat_size, out_hidden_size=1024, last_feat_dim=4096, p_dropout=0.5
    ):
        """
        :param input_flat_size:
        :param out_hidden_size:
        :param last_feat_dim:
        """
        super().__init__()
        self.h_layers = nn.Sequential(
            nn.Dropout(p=p_dropout),
            nn.Linear(input_flat_size, last_feat_dim),
            nn.ReLU(inplace=True),
            nn.Dropout(p=p_dropout),
            nn.Linear(last_feat_dim, out_hidden_size),
            nn.ReLU(inplace=True),
        )

    def forward(self, tensor_x):
        """
        :param x:
        """
        flat = tensor_x.view(tensor_x.shape[0], -1)
        out = self.h_layers(flat)
        return out



from domainlab.compos.vae.zoo_vae_builders_classif import (
    NodeVAEBuilderArg,
    NodeVAEBuilderImgAlex,
    NodeVAEBuilderImgConvBnPool,
    NodeVAEBuilderUser,
)
from domainlab.compos.vae.zoo_vae_builders_classif_topic import NodeVAEBuilderImgTopic


class VAEChainNodeGetter(object):
    """
    1. Hardcoded chain, each node use Scenario as request class
    2. Constructor takes parameters for VABuilder Subclasses
    3. heavy weight business objective is returned by selected node
    4. convert Scenario object to request object, so that class can be reused
    """

    def __init__(self, request, topic_dim=None):
        """ """
        self.request = request
        self.topic_dim = topic_dim

    def __call__(self):
        """
        1. construct the chain, filter out responsible node,
        create heavy-weight business object
        2. hard code seems to be the best solution
        """
        if self.topic_dim is not None:
            chain = NodeVAEBuilderImgTopic(None)
        else:
            chain = NodeVAEBuilderImgConvBnPool(None)
            chain = NodeVAEBuilderImgAlex(chain)
            chain = NodeVAEBuilderArg(chain)
            chain = NodeVAEBuilderUser(chain)
        node = chain.handle(self.request)
        return node



"""
test task for image size 224
"""
import os

from torchvision import transforms

from domainlab.tasks.task_folder_mk import mk_task_folder
from domainlab.tasks.utils_task import ImSize

path_this_file = os.path.dirname(os.path.realpath(__file__))


def addtask2chain(chain):
    """
    given a chain of responsibility for task selection, add another task into the chain
    """
    new_chain = mk_task_folder(
        extensions={"caltech": "jpg", "sun": "jpg", "labelme": "jpg"},
        list_str_y=["chair", "car"],
        dict_domain_folder_name2class={
            "caltech": {"auto": "car", "stuhl": "chair"},
            "sun": {"vehicle": "car", "sofa": "chair"},
            "labelme": {"drive": "car", "sit": "chair"},
        },
        dict_domain_img_trans={
            "caltech": transforms.Compose(
                [transforms.Resize((224, 224)), transforms.ToTensor()]
            ),
            "sun": transforms.Compose(
                [transforms.Resize((224, 224)), transforms.ToTensor()]
            ),
            "labelme": transforms.Compose(
                [transforms.Resize((224, 224)), transforms.ToTensor()]
            ),
        },
        img_trans_te=transforms.Compose(
            [transforms.Resize((224, 224)), transforms.ToTensor()]
        ),
        isize=ImSize(3, 224, 224),
        dict_domain2imgroot={
            "caltech": os.path.join(
                path_this_file,
                os.path.normpath("../"),
                os.path.normpath("zdata/vlcs_mini/caltech/"),
            ),
            "sun": os.path.join(
                path_this_file,
                os.path.normpath("../"),
                os.path.normpath("zdata/vlcs_mini/sun/"),
            ),
            "labelme": os.path.join(
                path_this_file,
                os.path.normpath("../"),
                os.path.normpath("zdata/vlcs_mini/labelme/"),
            ),
        },
        taskna="mini_vlcs",
        succ=chain,
    )
    return new_chain



"""Classification Performance"""
import numpy as np
import torch
from torchmetrics.classification import (
    AUC,
    AUROC,
    Accuracy,
    ConfusionMatrix,
    F1Score,
    Precision,
    Recall,
    Specificity,
)


class PerfMetricClassif:
    """Classification Performance metrics"""

    def __init__(self, num_classes, agg_precision_recall_f1="macro"):
        super().__init__()
        self.acc = Accuracy(num_classes=num_classes, average="micro")
        # NOTE: only micro aggregation make sense for acc
        self.precision = Precision(
            num_classes=num_classes, average=agg_precision_recall_f1
        )
        # Calculate the metric for each class separately, and average the
        # metrics across classes (with equal weights for each class).
        self.recall = Recall(num_classes=num_classes, average=agg_precision_recall_f1)
        self.f1_score = F1Score(
            num_classes=num_classes, average=agg_precision_recall_f1
        )
        # NOTE: auroc does nto support "micro" as aggregation
        self.auroc = AUROC(num_classes=num_classes, average=agg_precision_recall_f1)
        self.specificity = Specificity(
            num_classes=num_classes, average=agg_precision_recall_f1
        )
        self.confmat = ConfusionMatrix(num_classes=num_classes)

    def cal_metrics(self, model, loader_te, device):
        """
        :param model:
        :param loader_te:
        :param device: for final test, GPU can be used
        """
        self.acc.reset()
        self.precision.reset()
        self.recall.reset()
        self.f1_score.reset()
        self.auroc.reset()
        self.specificity.reset()
        self.confmat.reset()
        self.acc = self.acc.to(device)
        self.precision = self.precision.to(device)
        self.recall = self.recall.to(device)
        self.f1_score = self.f1_score.to(device)
        self.auroc = self.auroc.to(device)
        self.specificity = self.specificity.to(device)
        self.confmat = self.confmat.to(device)
        model_local = model.to(device)
        model_local.eval()
        with torch.no_grad():
            for _, (x_s, y_s, *_) in enumerate(loader_te):
                x_s, y_s = x_s.to(device), y_s.to(device)
                _, prob, _, *_ = model_local.infer_y_vpicn(x_s)
                _, target_label = torch.max(y_s, 1)
                # self.acc.update(pred_label, target_label)  # bug found by xinyuejohn
                # pred_label is (N,C) which we need to convert to (N)
                # prob is (N,C) which is fine for torchmetrics
                self.acc.update(prob, target_label)
                self.precision.update(prob, target_label)
                self.recall.update(prob, target_label)
                self.specificity.update(prob, target_label)
                self.f1_score.update(prob, target_label)
                self.auroc.update(prob, target_label)
                self.confmat.update(prob, target_label)

        acc_y = self.acc.compute()
        precision_y = self.precision.compute()
        recall_y = self.recall.compute()
        specificity_y = self.specificity.compute()
        f1_score_y = self.f1_score.compute()
        auroc_y = self.auroc.compute()
        confmat_y = self.confmat.compute()
        dict_metric = {
            "acc": acc_y,
            "precision": precision_y,
            "recall": recall_y,
            "specificity": specificity_y,
            "f1": f1_score_y,
            "auroc": auroc_y,
            "confmat": confmat_y,
        }
        keys = list(dict_metric)
        keys.remove("confmat")
        for key in keys:
            dict_metric[key] = dict_metric[key].cpu().numpy().sum()
        dict_metric["confmat"] = dict_metric["confmat"].cpu().numpy()
        return dict_metric



"""Classification Performance"""
import torch


class PerfClassif:
    """Classification Performance"""

    @classmethod
    def gen_fun_acc(cls, dim_target):
        """
        :param dim_target: class/domain label embeding dimension
        """

        def fun_acc(list_vec_preds, list_vec_labels):
            """
            :param list_vec_preds: list of batches
            """
            assert len(list_vec_preds) > 0
            correct_count = 0
            obs_count = 0
            for pred, label in zip(list_vec_preds, list_vec_labels):
                correct_count += torch.sum(
                    torch.sum(pred == label, dim=1) == dim_target
                )
                obs_count += pred.shape[0]  # batch size
            if isinstance(correct_count, int):
                acc = (correct_count) / obs_count
            else:
                acc = (correct_count.float()) / obs_count
            # AttributeError: 'int' object has no attribute 'float'
            # reason: batchsize is too big
            return acc

        return fun_acc

    @classmethod
    def cal_acc(cls, model, loader_te, device):
        """
        :param model:
        :param loader_te:
        :param device: for final test, GPU can be used
        """
        model.eval()
        model_local = model.to(device)
        fun_acc = cls.gen_fun_acc(model_local.dim_y)
        list_vec_preds, list_vec_labels = cls.get_list_pred_target(
            model_local, loader_te, device
        )
        accuracy_y = fun_acc(list_vec_preds, list_vec_labels)
        acc_y = accuracy_y.cpu().numpy().item()
        return acc_y

    @classmethod
    def get_list_pred_target(cls, model_local, loader_te, device):
        """
        isolate function to check if prediction persist each time loader is went through
        """
        list_vec_preds, list_vec_labels = [], []
        with torch.no_grad():
            for _, (x_s, y_s, *_) in enumerate(loader_te):
                x_s, y_s = x_s.to(device), y_s.to(device)
                pred, *_ = model_local.infer_y_vpicn(x_s)
                list_vec_preds.append(pred)
                list_vec_labels.append(y_s)
        return list_vec_preds, list_vec_labels



"""
choose devices
"""
import torch

from domainlab.utils.logger import Logger


def get_device(args):
    """
    choose devices
    """
    flag_no_cu = args.nocu
    flag_cuda = torch.cuda.is_available() and (not flag_no_cu)
    if args.device is None:
        device = torch.device("cuda" if flag_cuda else "cpu")
    else:
        device = torch.device("cuda:" + args.device if flag_cuda else "cpu")
    logger = Logger.get_logger()
    logger.info("")
    logger.info(f"using device: {str(device)}")
    logger.info("")
    return device



"""
build algorithm from API coded model with custom backbone
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderAPIModel(NodeAlgoBuilder):
    """
    build algorithm from API coded model with custom backbone
    """

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        args = exp.args
        device = get_device(args)
        model_sel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
        observer = ObVisitor(model_sel)
        trainer = TrainerChainNodeGetter(args.trainer)(default="hyperscheduler")
        return trainer, None, observer, device



from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter
from domainlab.utils.utils_cuda import get_device


def make_basic_trainer(class_name_model):
    """make_basic_trainer.
    :param class_name_model:
    """

    class NodeAlgoBuilderCustom(NodeAlgoBuilder):
        """NodeAlgoBuilderCustom."""

        def get_trainer(self, args):
            """
            chain of responsibility pattern for fetching trainer from commandline parsed arguments
            """
            trainer = TrainerChainNodeGetter(args.trainer)(default="basic")
            return trainer

        def _set_args(self, args, val_arg_na, prefix, argname):
            """_set_args.
            Since we could not hard code all possible strings in the argparser,
            we add attributes to args name space according to the user
            specification from the custom python file they provide

            the custom model the user wrote should have
            model.dict_net_module_na2arg_na with something like
            {"net1":"name1", "net2":"name2"}
            val_arg_na below will be filled with "name1" for instance

            python main_out.py –te_d=caltech –task=mini_vlcs –debug –bs=3
            –apath=examples/algos/demo_custom_model.py
            –model=custom –nname_argna2val net1 –nname_argna2val alexnet

            :param args: the namespace of command line arguemnts
            :param val_arg_na: the custom argument name the user specified
            :param prefix: nname or npath to be consistent with the rest of
            the package
            :param argname: nname_argna2val or "npath_argna2val", hard coded
            """
            if getattr(args, argname) is None:
                setattr(args, prefix + val_arg_na, None)
                return
            list_args = getattr(args, argname)
            ind = list_args.index(val_arg_na)
            if ind + 1 >= len(list_args):  # list of args always even length
                raise RuntimeError(
                    f"\n nname_argna2val or npath_argna2val should \
                                   \n always be specified in pairs instead of \
                                   odd number:\
                                   \n {str(list_args)}"
                )
            val = list_args[ind + 1]
            # add attributes to namespaces args, the attributes are provided by
            # user in the custom model file
            setattr(args, prefix + val_arg_na, val)

        def set_nets_from_dictionary(self, args, task, model):
            """set_nets_from_dictionary.
            the custom model the user wrote should have
            model.dict_net_module_na2arg_na with something like
            {"net1":"name1", "net2":"name2"}
            python main_out.py –te_d=caltech –task=mini_vlcs –debug –bs=3
            –apath=examples/algos/demo_custom_model.py
            –model=custom –nname_argna2val net1 –nname_argna2val alexnet
            """
            for key_module_na, val_arg_na in model.dict_net_module_na2arg_na.items():
                #
                if args.nname_argna2val is None and args.npath_argna2val is None:
                    raise RuntimeError(
                        "either specify nname_argna2val or \
                                        npath_argna2val"
                    )
                self._set_args(args, val_arg_na, "nname", "nname_argna2val")
                self._set_args(args, val_arg_na, "npath", "npath_argna2val")
                #
                builder = FeatExtractNNBuilderChainNodeGetter(
                    args,
                    arg_name_of_net="nname" + val_arg_na,
                    arg_path_of_net="npath" + val_arg_na,
                )()
                net = builder.init_business(
                    flag_pretrain=True,
                    dim_out=task.dim_y,
                    remove_last_layer=False,
                    args=args,
                    isize=(task.isize.i_c, task.isize.i_h, task.isize.i_w),
                )
                model.add_module("%s" % (key_module_na), net)

        def init_business(self, exp):
            """
            return trainer, model, observer
            """
            task = exp.task
            args = exp.args
            device = get_device(args)
            model_sel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
            observer = ObVisitor(model_sel)
            model = class_name_model(list_str_y=task.list_str_y)
            model = self.init_next_model(model, exp)
            self.set_nets_from_dictionary(args, task, model)
            trainer = self.get_trainer(args)
            trainer.init_business(model, task, observer, device, args)
            return trainer, model, observer, device

    return NodeAlgoBuilderCustom



"""
args for matchdg
"""


def add_args2parser_matchdg(parser):
    """
    args for matchdg
    """
    # parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tau", type=float, default=0.05, help="factor to magnify cosine similarity"
    )
    parser.add_argument(
        "--epos_per_match_update",
        type=int,
        default=5,
        help="Number of epochs before updating the match tensor",
    )
    parser.add_argument(
        "--epochs_ctr", type=int, default=1, help="Total number of epochs for ctr"
    )
    # args = parser.parse_args("")
    # return args
    return parser



"""
interface for observer + visitor
"""
import abc


class AObVisitor(metaclass=abc.ABCMeta):
    """
    Observer + Visitor pattern for model selection
    """

    def __init__(self):
        self.task = None
        self.device = None
        self.epo_te = None
        self.str_msel = None
        self.keep_model = None
        self.loader_te = None
        self.loader_tr = None
        self.loader_val = None

    def set_task(self, task, args, device):
        """
        couple observer with task
        """
        self.task = task
        self.device = device
        self.epo_te = args.epo_te
        self.str_msel = args.msel  # FIXME: consistent with self.model_sel?
        self.keep_model = args.keep_model
        self.loader_te = self.task.loader_te
        self.loader_tr = self.task.loader_tr
        # Note loader_tr behaves/inherit different properties than loader_te
        self.loader_val = self.task.loader_val

    @abc.abstractmethod
    def update(self, epoch) -> bool:
        """
        return True/False whether the trainer should stop training
        """

    @abc.abstractmethod
    def accept(self, trainer):
        """
        accept invitation as a visitor
        """

    @abc.abstractmethod
    def after_all(self):
        """
        After training is done
        """

    @abc.abstractmethod
    def clean_up(self):
        """
        to be called by a decorator
        """

    def reset(self):
        """
        reset observer via reset model selector
        """
        self.model_sel.reset()



"""
This adaptor couples intensively with the  heavy-weight model class
The model class can be refactored, we do want to use the trained old-version model,
which we only need to change this adaptor class.
"""


class AdaptorReconVAEXYD:
    """
    This adaptor couples intensively with the  heavy-weight model class
    The model class can be refactored, we do want to use the trained old-version model,
    which we only need to change this adaptor class.
    """

    def __init__(self, model):
        self.model = model

    def cal_latent(self, x):
        """
        This method won't be redundant as it will be used several times,
        and most importantly, it **couples** with the models encoder attribute,
        so we only need to change this one time.
        Suppose if model class changes, to use old trained models(we can not retrain them),
        we only need to change this method.
        :param x:
        """
        q_zd, _, q_zx, _, q_zy, _ = self.model.encoder(x)
        return q_zd, q_zx, q_zy

    def recon_ydx(self, zy, zd, zx, x):
        """
        1. The order of concatnation plays a vital role!
        2. This method won't be redundant as it will be used several times,
        and most importantly, it **couples** with the models encoder attribute,
        so we only need to change this one time.
        Suppose if model class changes, to use old trained models(we can not retrain them),
        we only need to change this method.
        """
        z_concat = self.model.decoder.concat_ydx(zy, zd, zx)
        _, x_mean, _ = self.model.decoder(z_concat, x)
        return x_mean

    def cal_prior_zy(self, vec_y):
        """ """
        p_zy = self.model.net_p_zy(vec_y)
        return p_zy

    def cal_prior_zd(self, vec_d):
        """ """
        p_zd = self.model.net_p_zd(vec_d)
        return p_zd



from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.utils.flows_gen_img_model import fun_gen
from domainlab.utils.logger import Logger


class ObVisitorGen(ObVisitor):
    """
    For Generative Models
    """

    def after_all(self):
        super().after_all()
        logger = Logger.get_logger()
        logger.info("generating images for final model at last epoch")
        fun_gen(
            subfolder_na=self.host_trainer.model.visitor.model_name + "final",
            args=self.host_trainer.aconf,
            node=self.host_trainer.task,
            model=self.host_trainer.model,
            device=self.device,
        )

        logger.info("generating images for oracle model")
        model_or = self.host_trainer.model.load(
            "oracle"
        )  # @FIXME: name "oracle is a strong dependency
        model_or = model_or.to(self.device)
        model_or.eval()
        fun_gen(
            subfolder_na=self.host_trainer.model.visitor.model_name + "oracle",
            args=self.host_trainer.aconf,
            node=self.host_trainer.task,
            model=model_or,
            device=self.device,
        )
        logger.info("generating images for selected model")
        model_ld = self.host_trainer.model.load()
        model_ld = model_ld.to(self.device)
        model_ld.eval()
        fun_gen(
            subfolder_na=self.host_trainer.model.visitor.model_name + "selected",
            args=self.host_trainer.aconf,
            node=self.host_trainer.task,
            model=model_ld,
            device=self.device,
        )



import torch.nn as nn
from torchvision import models as torchvisionmodels

from domainlab.compos.nn_zoo.nn import LayerId
from domainlab.compos.nn_zoo.nn_torchvision import NetTorchVisionBase
from domainlab.utils.logger import Logger


class AlexNetBase(NetTorchVisionBase):
    """
    .. code-block:: python

       AlexNet(
       (features): Sequential(
           (0): Conv2d(3, 64, kernel_size=(11, 11), stride=(4, 4), padding=(2, 2))
           (1): ReLU(inplace=True)
           (2): MaxPool2d(kernel_size=3, stride=2, padding=0, dilation=1, ceil_mode=False)
           (3): Conv2d(64, 192, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2))
           (4): ReLU(inplace=True)
           (5): MaxPool2d(kernel_size=3, stride=2, padding=0, dilation=1, ceil_mode=False)
           (6): Conv2d(192, 384, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
           (7): ReLU(inplace=True)
           (8): Conv2d(384, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
           (9): ReLU(inplace=True)
           (10): Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
           (11): ReLU(inplace=True)
           (12): MaxPool2d(kernel_size=3, stride=2, padding=0, dilation=1, ceil_mode=False)
       )
       (avgpool): AdaptiveAvgPool2d(output_size=(6, 6))
       (classifier): Sequential(
           (0): Dropout(p=0.5, inplace=False)
           (1): Linear(in_features=9216, out_features=4096, bias=True)
           (2): ReLU(inplace=True)
           (3): Dropout(p=0.5, inplace=False)
           (4): Linear(in_features=4096, out_features=4096, bias=True)
           (5): ReLU(inplace=True)
           (6): Linear(in_features=4096, out_features=7, bias=True)
       )
       )
    """

    def fetch_net(self, flag_pretrain):
        self.net_torchvision = torchvisionmodels.alexnet(pretrained=flag_pretrain)


class Alex4DeepAll(AlexNetBase):
    """
    change the last layer output of AlexNet to the dimension of the
    """

    def __init__(self, flag_pretrain, dim_y):
        super().__init__(flag_pretrain)
        if self.net_torchvision.classifier[6].out_features != dim_y:
            logger = Logger.get_logger()
            logger.info(
                f"original alex net out dim "
                f"{self.net_torchvision.classifier[6].out_features}"
            )
            num_ftrs = self.net_torchvision.classifier[6].in_features
            self.net_torchvision.classifier[6] = nn.Linear(num_ftrs, dim_y)
            logger.info(f"re-initialized to {dim_y}")


class AlexNetNoLastLayer(AlexNetBase):
    """
    Change the last layer of AlexNet with identity layer,
    the classifier from VAE can then have the same layer depth as erm
    model so it is fair for comparison
    """

    def __init__(self, flag_pretrain):
        super().__init__(flag_pretrain)
        self.net_torchvision.classifier[6] = LayerId()



"""
DIVA
"""
import torch
from torch.nn import functional as F

from domainlab import g_inst_component_loss_agg, g_str_cross_entropy_agg
from domainlab.models.model_vae_xyd_classif import VAEXYDClassif
from domainlab.utils.utils_class import store_args


def mk_diva(parent_class=VAEXYDClassif):
    """
    Instantiate a domain invariant variational autoencoder (DIVA) with arbitrary task loss.

    Details:
        This method is creating a generative model based on a variational autoencoder, which can
        reconstruct the input images. Here for, three different encoders with latent variables are
        trained, each representing a latent subspace for the domain, class and residual features
        information, respectively. The latent subspaces serve for disentangling the respective
        sources of variation. To reconstruct the input image, the three latent variables are fed
        into a decoder.
        Additionally, two classifiers are trained, which predict the domain and the class label.
        For more details, see:
        Ilse, Maximilian, et al. "Diva: Domain invariant variational autoencoders."
        Medical Imaging with Deep Learning. PMLR, 2020.

    Args:
        parent_class: Class object determining the task type. Defaults to VAEXYDClassif.

    Returns:
        ModelDIVA: model inheriting from parent class.

    Input Parameters:
        zd_dim: size of latent space for domain-specific information,
        zy_dim: size of latent space for class-specific information,
        zx_dim: size of latent space for residual variance,
        chain_node_builder: creates the neural network specified by the user; object of the class
        "VAEChainNodeGetter" (see domainlab/compos/vae/utils_request_chain_builder.py)
        being initialized by entering a user request,
        list_str_y: list of labels,
        list_d_tr: list of training domains,
        gamma_d: weighting term for d classifier,
        gamma_y: weighting term for y classifier,
        beta_d: weighting term for domain encoder,
        beta_x: weighting term for residual variation encoder,
        beta_y: weighting term for class encoder

    Usage:
        For a concrete example, see:
        https://github.com/marrlab/DomainLab/blob/master/tests/test_mk_exp_diva.py
    """

    class ModelDIVA(parent_class):
        """
        DIVA
        """

        @store_args
        def __init__(
            self,
            chain_node_builder,
            zd_dim,
            zy_dim,
            zx_dim,
            list_str_y,
            list_d_tr,
            gamma_d,
            gamma_y,
            beta_d,
            beta_x,
            beta_y,
            multiplier_recon=1.0,
        ):
            """
            gamma: classification loss coefficient
            """
            super().__init__(chain_node_builder, zd_dim, zy_dim, zx_dim, list_str_y)
            self.list_d_tr = list_d_tr
            self.dim_d_tr = len(self.list_d_tr)
            if self.zd_dim > 0:
                self.add_module(
                    "net_p_zd",
                    self.chain_node_builder.construct_cond_prior(
                        self.dim_d_tr, self.zd_dim
                    ),
                )
                self.add_module(
                    "net_classif_d",
                    self.chain_node_builder.construct_classifier(
                        self.zd_dim, self.dim_d_tr
                    ),
                )

        def hyper_update(self, epoch, fun_scheduler):
            """hyper_update.

            :param epoch:
            :param fun_scheduler:
            """
            dict_rst = fun_scheduler(epoch)
            self.beta_d = dict_rst["beta_d"]
            self.beta_y = dict_rst["beta_y"]
            self.beta_x = dict_rst["beta_x"]

        def hyper_init(self, functor_scheduler):
            """
            initiate a scheduler object via class name and things inside this model

            :param functor_scheduler: the class name of the scheduler
            """
            return functor_scheduler(
                trainer=None, beta_d=self.beta_d, beta_y=self.beta_y, beta_x=self.beta_x
            )

        def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
            q_zd, zd_q, q_zx, zx_q, q_zy, zy_q = self.encoder(tensor_x)
            logit_d = self.net_classif_d(zd_q)

            batch_size = zd_q.shape[0]
            device = zd_q.device

            p_zx = self.init_p_zx4batch(batch_size, device)
            p_zy = self.net_p_zy(tensor_y)
            p_zd = self.net_p_zd(tensor_d)

            z_concat = self.decoder.concat_ydx(zy_q, zd_q, zx_q)
            loss_recon_x, _, _ = self.decoder(z_concat, tensor_x)

            zd_p_minus_zd_q = g_inst_component_loss_agg(
                p_zd.log_prob(zd_q) - q_zd.log_prob(zd_q), 1
            )
            # without aggregation, shape is [batchsize, zd_dim]
            zx_p_minus_zx_q = torch.zeros_like(zd_p_minus_zd_q)
            if self.zx_dim > 0:
                # torch.sum will return 0 for empty tensor,
                # torch.mean will return nan
                zx_p_minus_zx_q = g_inst_component_loss_agg(
                    p_zx.log_prob(zx_q) - q_zx.log_prob(zx_q), 1
                )

            zy_p_minus_zy_q = g_inst_component_loss_agg(
                p_zy.log_prob(zy_q) - q_zy.log_prob(zy_q), 1
            )

            _, d_target = tensor_d.max(dim=1)
            lc_d = F.cross_entropy(logit_d, d_target, reduction=g_str_cross_entropy_agg)

            return [
                loss_recon_x,
                zd_p_minus_zd_q,
                zx_p_minus_zx_q,
                zy_p_minus_zy_q,
                lc_d,
            ], [
                self.multiplier_recon,
                -self.beta_d,
                -self.beta_x,
                -self.beta_y,
                self.gamma_d,
            ]

    return ModelDIVA



"""
https://github.com/pytorch/vision/blob/bb5af1d77658133af8be8c9b1a13139722315c3a/torchvision/datasets/folder.py#L93
https://pytorch.org/vision/stable/_modules/torchvision/datasets/folder.html#DatasetFolder.fetch_img_paths
"""
import os
import sys
import warnings
from typing import Any, Tuple

from torchvision.datasets import DatasetFolder

from domainlab.utils.logger import Logger


def has_file_allowed_extension(filename: str, extensions: Tuple[str, ...]) -> bool:
    """
    Checks if a file is an allowed extension.
    Args:
    filename (string): path to a file
    extensions (tuple of strings): extensions to consider (lowercase)
    Returns: bool: True if the filename ends with one of given extensions
    """
    return filename.lower().endswith(extensions)


def fetch_img_paths(path_dir, class_to_idx, extensions=None, is_valid_file=None):
    """
    :param path_dir: path to fetch images in string format
    :param class_to_idx: given list of strings as class names
    {classes[i]: i for i in range(len(classes))}
    :param extensions: file extensions in fstring format
    :param is_valid_file: user provided function to check if the file is valid or not
    :return : list_tuple_path_cls_ind: list of tuple, (path of file, class index)
    """
    list_tuple_path_cls_ind = []
    path_dir = os.path.expanduser(path_dir)
    # since this function is only called by the class below, which now ensures that
    # extensions xor is_valid_file is not None, this check cannot be triggered
    # if not ((extensions is None) ^ (is_valid_file is None)):
    #     raise ValueError(
    #         "Both extensions and is_valid_file cannot be None or not None at the same time")
    if extensions is not None:

        def functor_is_valid_file(filena):
            return has_file_allowed_extension(filena, extensions)

        is_valid_file = functor_is_valid_file
    for target in sorted(class_to_idx.keys()):
        apath = os.path.join(path_dir, target)
        if not os.path.isdir(apath):
            continue
        for root, _, fnames in sorted(os.walk(apath, followlinks=True)):
            for fname in sorted(fnames):
                path_file = os.path.join(root, fname)
                if is_valid_file(path_file):
                    item = (path_file, class_to_idx[target])
                    list_tuple_path_cls_ind.append(item)  # @FIXME
    return list_tuple_path_cls_ind


class DsetSubFolder(DatasetFolder):
    """
    Only use user provided class names, ignore the other subfolders
    :param list_class_dir: list of class directories to use as classes
    """

    def __init__(
        self,
        root,
        loader,
        list_class_dir,
        extensions=None,
        transform=None,
        target_transform=None,
        is_valid_file=None,
    ):
        self.list_class_dir = list_class_dir

        if is_valid_file is not None and extensions is not None:
            raise ValueError(
                "Both extensions and is_valid_file cannot be not None at the same time"
            )

        if is_valid_file is None and extensions is None:
            # setting default extensions
            extensions = ("jpg", "jpeg", "png")
            logger = Logger.get_logger()
            logger.warn("no user provided extensions, set to be jpg, jpeg, png")
            warnings.warn("no user provided extensions, set to be jpg, jpeg, png")

        super().__init__(
            root,
            loader,
            extensions=extensions,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
        )
        classes, class_to_idx = self._find_classes(self.root)
        samples = fetch_img_paths(self.root, class_to_idx, extensions, is_valid_file)
        self.classes = classes
        self.class_to_idx = class_to_idx
        self.samples = samples
        self.targets = [s[1] for s in samples]
        self.path2imgs = [s[0] for s in samples]

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target, (path,)

    def _find_classes(self, mdir):
        """
        Finds the class folders in a dataset.
        Args:
            mdir (string): Root mdirectory path.
        Returns:
            tuple: (classes, class_to_idx) where classes are relative to (mdir),
            and class_to_idx is a dictionary.
        Ensures:
            No class is a submdirectory of another.
        """
        logger = Logger.get_logger()
        if sys.version_info >= (3, 5):
            # Faster and available in Python 3.5 and above
            list_subfolders = [subfolder.name for subfolder in list(os.scandir(mdir))]
            logger.info(f"list of subfolders {list_subfolders}")
            classes = [
                d.name
                for d in os.scandir(mdir)
                if d.is_dir() and d.name in self.list_class_dir
            ]
        else:
            classes = [
                d
                for d in os.listdir(mdir)
                if os.path.isdir(os.path.join(mdir, d)) and d in self.list_class_dir
            ]
        flag_user_input_classes_in_folder = set(self.list_class_dir) <= set(classes)
        if not flag_user_input_classes_in_folder:
            logger.info(f"user provided class names: {self.list_class_dir}")
            logger.info(f"subfolder names from folder: {mdir} {classes}")
            raise RuntimeError(
                "user provided class names does not match the subfolder names"
            )
        classes.sort()
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        return classes, class_to_idx



"""
decoder which takes concatenated latent representation
"""
import torch
import torch.nn as nn


class DecoderConcatLatentFcReshapeConv(nn.Module):
    """
    Latent vector re-arranged to image-size directly,
    then convolute to get the textures of the original image
    """

    def __init__(
        self,
        z_dim,
        i_c,
        i_h,
        i_w,
        cls_fun_nll_p_x,
        net_fc_z2flat_img,
        net_conv,
        net_p_x_mean,
        net_p_x_log_var,
    ):
        """
        :param z_dim:
        :param list_im_chw: [channel, height, width]
        """
        self.z_dim = z_dim
        self.cls_fun_nll_p_x = cls_fun_nll_p_x
        self.list_im_chw = [i_c, i_h, i_w]
        super().__init__()
        self.add_module("net_fc_z2flat_img", net_fc_z2flat_img)
        self.add_module("net_conv", net_conv)
        self.add_module("net_p_x_mean", net_p_x_mean)
        self.add_module("net_p_x_log_var", net_p_x_log_var)

    def cal_p_x_pars_loc_scale(self, vec_z):
        """compute mean and variance of each pixel
        :param z:
        """
        h_flat = self.net_fc_z2flat_img(vec_z)
        # reshape to image
        h_img = h_flat.view(
            -1, self.list_im_chw[0], self.list_im_chw[1], self.list_im_chw[2]
        )
        h_img_conv = self.net_conv(h_img)
        # pixel must be positive: enforced by sigmoid activation
        x_mean = self.net_p_x_mean(h_img_conv)  # .view(-1, np.prod(self.list_im_chw))
        # remove the saturated part of sigmoid
        x_mean = torch.clamp(x_mean, min=0.0 + 1.0 / 512.0, max=1.0 - 1.0 / 512.0)
        # negative values
        x_logvar = self.net_p_x_log_var(
            h_img_conv
        )  # .view(-1, np.prod(self.list_im_chw))
        return x_mean, x_logvar

    def concat_ydx(self, zy, zd, zx):
        z_concat = torch.cat((zy, zd, zx), 1)
        return z_concat

    def concat_ytdx(self, zy, topic, zd, zx):
        z_concat = torch.cat((zy, topic, zd, zx), 1)
        return z_concat

    def forward(self, z, img):
        """
        :param z:
        :param img:
        """
        x_mean, x_logvar = self.cal_p_x_pars_loc_scale(z)
        nll = self.cls_fun_nll_p_x()(img, x_mean, x_logvar)
        return nll, x_mean, x_logvar



"Classifier"
import torch
import torch.nn as nn
from torch.nn import functional as F


class ClassifDropoutReluLinear(nn.Module):
    """first apply dropout, then relu, then linearly fully connected, without activation"""

    def __init__(self, z_dim, target_dim):
        """
        :param z_dim:
        :param target_dim:
        """
        super().__init__()
        self.op_drop = nn.Dropout()
        self.op_linear = nn.Linear(z_dim, target_dim)
        torch.nn.init.xavier_uniform_(self.op_linear.weight)
        self.op_linear.bias.data.zero_()

    def forward(self, z_vec):
        """
        :param z_vec:
        """
        hidden = F.relu(self.op_drop(z_vec))
        logit = self.op_linear(hidden)
        return logit



"""
Abstract Model Selection
"""

import abc


class AMSel(metaclass=abc.ABCMeta):
    """
    Abstract Model Selection
    """

    def __init__(self):
        """
        trainer and tr_observer
        """
        self.trainer = None
        self._tr_obs = None
        self.msel = None
        self._max_es = None

    def reset(self):
        """
        reset observer via reset model selector
        """
        if self.msel is not None:
            self.msel.reset()

    @property
    def tr_obs(self):
        """
        the observer from trainer
        """
        return self._tr_obs

    @property
    def max_es(self):
        """
        maximum early stop
        """
        if self._max_es is not None:
            return self._max_es
        if self.msel is not None:
            return self.msel.max_es
        return self._max_es

    def accept(self, trainer, tr_obs):
        """
        Visitor pattern to trainer
        accept trainer and tr_observer
        """
        self.trainer = trainer
        self._tr_obs = tr_obs
        if self.msel is not None:
            self.msel.accept(trainer, tr_obs)

    @abc.abstractmethod
    def update(self, clear_counter=False):
        """
        observer + visitor pattern to trainer
        if the best model should be updated
        return boolean
        """

    def if_stop(self):
        """
        check if trainer should stop
        return boolean
        """
        # NOTE: since if_stop is not abstract, one has to
        # be careful to always override it in child class
        # only if the child class has a decorator which will
        # dispatched.
        if self.msel is not None:
            return self.msel.if_stop()
        raise NotImplementedError

    @property
    def best_val_acc(self):
        """
        decoratee best val acc
        """
        if self.msel is not None:
            return self.msel.best_val_acc
        return -1

    @property
    def best_te_metric(self):
        """
        decoratee best test metric
        """
        if self.msel is not None:
            return self.msel.best_te_metric
        return -1

    @property
    def sel_model_te_acc(self):
        """
        the selected model test accuaracy
        """
        if self.msel is not None:
            return self.msel.sel_model_te_acc
        return -1



def override_interface(interface_class):
    """overrides.
    :param interface_class:  the interface class name, always specify this
    explicitly as otherwise interface_class is going to be the nearest
    function it decorate, and argument "method2override" of returned
    function "overrider" accept will be the current child class

    .. code-block:: python

       class BaseClass()
       class Child(BaseClass):
           @overrides(BaseClass)
           def fun(self):
               pass
    """

    def overrider(method2override):
        """overrider.

        :param method:  the method to be decorated, method signature will be
        returned intactly
        """
        assert method2override.__name__ in dir(interface_class)
        return method2override

    return overrider



"""
Pytorch image: i_channel, i_h, i_w
Location-Scale Encoder: SoftPlus
"""
import torch
import torch.distributions as dist
import torch.nn as nn

from domainlab.compos.nn_zoo.net_conv_conv_bn_pool_2 import mk_conv_bn_relu_pool
from domainlab.compos.nn_zoo.nn import DenseNet
from domainlab.compos.utils_conv_get_flat_dim import get_flat_dim


class LSEncoderConvBnReluPool(nn.Module):
    """Location-Scale Encoder with Convolution,
    Batch Normalization, Relu and Pooling.
    Softplus for scale
    """

    def __init__(self, z_dim: int, i_channel, i_h, i_w, conv_stride):
        """
        :param z_dim:
        nn.Sequential allows output dim to be zero.
        So z_dim here can be set to be zero
        :param i_channel:
        :param i_h:
        :param i_w:
        :param conv_stride:
        """
        super().__init__()
        self.i_channel = i_channel
        self.i_h = i_h
        self.i_w = i_w

        self.conv = mk_conv_bn_relu_pool(self.i_channel, conv_stride=conv_stride)
        # conv-bn-relu-pool-conv-bn-relu-pool(no activation)
        self.flat_dim = get_flat_dim(self.conv, i_channel, i_h, i_w)
        self.fc_loc = nn.Sequential(nn.Linear(self.flat_dim, z_dim))
        self.fc_scale = nn.Sequential(
            nn.Linear(self.flat_dim, z_dim), nn.Softplus()
        )  # for scale calculation

        # initialization
        torch.nn.init.xavier_uniform_(self.fc_loc[0].weight)
        self.fc_loc[0].bias.data.zero_()
        torch.nn.init.xavier_uniform_(self.fc_scale[0].weight)
        self.fc_scale[0].bias.data.zero_()

    def forward(self, img):
        """.
        :param img:
        """
        hidden = self.conv(img)
        hidden = hidden.view(-1, self.flat_dim)
        zd_q_loc = self.fc_loc(hidden)
        zd_q_scale = self.fc_scale(hidden) + 1e-7
        q_zd = dist.Normal(zd_q_loc, zd_q_scale)
        zd_q = q_zd.rsample()  # Reparameterization trick
        return q_zd, zd_q


class LSEncoderLinear(nn.Module):
    """
    Location-Scale Encoder with DenseNet as feature extractor
    Softplus for scale
    """

    def __init__(self, z_dim, dim_input):
        """
        :param z_dim:
        nn.Sequential allows output dim to be zero.
        So z_dim here can be set to be zero
        """
        super().__init__()
        self.fc_loc = nn.Sequential(nn.Linear(dim_input, z_dim))
        self.fc_scale = nn.Sequential(
            nn.Linear(dim_input, z_dim), nn.Softplus()
        )  # for scale calculation

        # initialization
        torch.nn.init.xavier_uniform_(self.fc_loc[0].weight)
        self.fc_loc[0].bias.data.zero_()
        torch.nn.init.xavier_uniform_(self.fc_scale[0].weight)
        self.fc_scale[0].bias.data.zero_()

    def forward(self, hidden):
        """.
        :param hidden:
        """
        zd_q_loc = self.fc_loc(hidden)
        zd_q_scale = self.fc_scale(hidden) + 1e-7
        q_zd = dist.Normal(zd_q_loc, zd_q_scale)
        zd_q = q_zd.rsample()  # Reparameterization trick
        return q_zd, zd_q



"""
Adaptor is vital for data generation so this class can be decoupled from model class.
The model class can be refactored, we do want to use the trained old-version model, which we
only need to change adaptor class.
"""
import torch

from domainlab.compos.vae.c_vae_adaptor_model_recon import AdaptorReconVAEXYD


class ReconVAEXYD:
    """
    Adaptor is vital for data generation so this class can be decoupled from model class.
    The model class can be refactored, we do want to use the trained old-version model, which we
    only need to change adaptor class.
    """

    def __init__(self, model, na_adaptor=AdaptorReconVAEXYD):
        self.model = model
        self.adaptor = na_adaptor(self.model)

    def recon(
        self,
        x,
        vec_y=None,
        vec_d=None,
        sample_p_zy=False,
        sample_p_zd=False,
        scalar_zx2fill=None,
    ):
        """
        common function
        """
        str_type = ""

        with torch.no_grad():
            q_zd, q_zx, q_zy = self.adaptor.cal_latent(x)

            if vec_d is None:
                recon_zd = q_zd.mean
                str_type = "zd_q"
            else:
                p_zd = self.adaptor.cal_prior_zd(vec_d)
                recon_zd = p_zd.mean
                str_type = "zd_p"
                if sample_p_zd:
                    recon_zd = p_zd.rsample()
                    str_type = "_".join([str_type, "sample__"])

            zx_loc_q = q_zx.mean
            if scalar_zx2fill is not None:
                recon_zx = torch.zeros_like(zx_loc_q)
                recon_zx = recon_zx.fill_(scalar_zx2fill)
                str_type = "_".join(
                    [str_type, "__fill_zx_", str(scalar_zx2fill), "___"]
                )
            else:
                recon_zx = zx_loc_q
                str_type = "_".join([str_type, "__zx_q__"])

            if vec_y is not None:
                p_zy = self.adaptor.cal_prior_zy(vec_y)
                recon_zy = p_zy.mean
                str_type = "_".join([str_type, "__zy_p__"])
                if sample_p_zy:
                    recon_zy = p_zy.rsample()
                    str_type = "_".join([str_type, "sample__"])
            else:
                recon_zy = q_zy.mean
                str_type = "_".join([str_type, "zy_q__"])
            img_recon = self.adaptor.recon_ydx(recon_zy, recon_zd, recon_zx, x)
            return img_recon, str_type

    def recon_cf(
        self,
        x,
        na_cf,
        dim_cf,
        device,
        vec_y=None,
        vec_d=None,
        zx2fill=None,
        sample_p_zy=False,
        sample_p_zd=False,
    ):
        """
        Countefactual reconstruction:
            :param na_cf: name of counterfactual, 'y' or 'd'
            :param dim_cf: dimension of counterfactual factor
        """
        list_recon_cf = []
        batch_size = x.shape[0]
        str_type = None
        # Counterfactual image generation
        for i in range(dim_cf):
            label_cf = torch.zeros(batch_size, dim_cf).to(device)
            label_cf[:, i] = 1
            if na_cf == "y":
                img_recon_cf, str_type = self.recon(
                    x, label_cf, vec_d, sample_p_zy, sample_p_zd, zx2fill
                )
            elif na_cf == "d":
                img_recon_cf, str_type = self.recon(
                    x, vec_y, label_cf, sample_p_zy, sample_p_zd, zx2fill
                )
            else:
                raise RuntimeError(
                    "counterfactual image generation can only be 'y' or 'd'"
                )
            list_recon_cf.append(img_recon_cf)
        return list_recon_cf, str_type



"""
basic trainer
"""
import math
from operator import add

import torch

from domainlab import g_tensor_batch_agg
from domainlab.algos.trainers.a_trainer import AbstractTrainer, mk_opt


def list_divide(list_val, scalar):
    return [ele / scalar for ele in list_val]


class TrainerBasic(AbstractTrainer):
    """
    basic trainer
    """

    def before_tr(self):
        """
        check the performance of randomly initialized weight
        """
        self.model.evaluate(self.loader_te, self.device)

    def before_epoch(self):
        """
        set model to train mode
        initialize some member variables
        """
        self.model.train()
        self.counter_batch = 0.0
        self.epo_loss_tr = 0
        self.epo_reg_loss_tr = [0.0 for _ in range(10)]
        self.epo_task_loss_tr = 0

    def tr_epoch(self, epoch):
        self.before_epoch()
        for ind_batch, (tensor_x, tensor_y, tensor_d, *others) in enumerate(
            self.loader_tr
        ):
            self.tr_batch(tensor_x, tensor_y, tensor_d, others, ind_batch, epoch)
        return self.after_epoch(epoch)

    def after_epoch(self, epoch):
        """
        observer collect information
        """
        self.epo_loss_tr /= self.counter_batch
        self.epo_task_loss_tr /= self.counter_batch
        self.epo_reg_loss_tr = list_divide(self.epo_reg_loss_tr, self.counter_batch)
        assert self.epo_loss_tr is not None
        assert not math.isnan(self.epo_loss_tr)
        flag_stop = self.observer.update(epoch)  # notify observer
        assert flag_stop is not None
        return flag_stop

    def log_loss(self, list_b_reg_loss, loss_task, loss):
        """
        just for logging the self.epo_reg_loss_tr
        """
        self.epo_task_loss_tr += loss_task.sum().detach().item()
        #
        list_b_reg_loss_sumed = [ele.sum().detach().item() for ele in list_b_reg_loss]
        self.epo_reg_loss_tr = list(
            map(add, self.epo_reg_loss_tr, list_b_reg_loss_sumed)
        )
        self.epo_loss_tr += loss.detach().item()

    def tr_batch(self, tensor_x, tensor_y, tensor_d, others, ind_batch, epoch):
        """
        optimize neural network one step upon a mini-batch of data
        """
        self.before_batch(epoch, ind_batch)
        tensor_x, tensor_y, tensor_d = (
            tensor_x.to(self.device),
            tensor_y.to(self.device),
            tensor_d.to(self.device),
        )
        self.optimizer.zero_grad()
        loss = self.cal_loss(tensor_x, tensor_y, tensor_d, others)
        loss.backward()
        self.optimizer.step()
        self.after_batch(epoch, ind_batch)
        self.counter_batch += 1

    def cal_loss(self, tensor_x, tensor_y, tensor_d, others):
        """
        so that user api can use trainer.cal_loss to train
        """
        loss_task = self.model.cal_task_loss(tensor_x, tensor_y)

        list_reg_tr_batch, list_mu_tr = self.cal_reg_loss(
            tensor_x, tensor_y, tensor_d, others
        )
        tensor_batch_reg_loss_penalized = self.model.list_inner_product(
            list_reg_tr_batch, list_mu_tr
        )
        assert len(tensor_batch_reg_loss_penalized.shape) == 1
        loss_erm_agg = g_tensor_batch_agg(loss_task)
        loss_reg_penalized_agg = g_tensor_batch_agg(tensor_batch_reg_loss_penalized)
        loss_penalized = (
            self.model.multiplier4task_loss * loss_erm_agg + loss_reg_penalized_agg
        )
        self.log_loss(list_reg_tr_batch, loss_task, loss_penalized)
        return loss_penalized



import torch
import torch.nn as nn

from domainlab.compos.vae.compos.encoder import LSEncoderLinear
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter


class EncoderSandwichTopicImg2Zd(nn.Module):
    """
    sandwich encoder: (img, s)->zd
    """

    def __init__(self, zd_dim, isize, num_topics, img_h_dim, args):
        """
        num_topics, img_h_dim: (img->h_img, topic->h_topic)-> q_zd
        :param img_h_dim: (img->h_img, topic->h_topic)-> q_zd
        the dimension to concatenate with topic vector to infer z_d
        """
        super().__init__()
        self.zd_dim = zd_dim
        self.img_h_dim = img_h_dim

        net_builder = FeatExtractNNBuilderChainNodeGetter(
            args=args,
            arg_name_of_net="nname_encoder_sandwich_x2h4zd",
            arg_path_of_net="npath_encoder_sandwich_x2h4zd",
        )()

        # image->h_img
        self.add_module(
            "layer_img2h4zd",
            net_builder.init_business(
                dim_out=self.img_h_dim,
                flag_pretrain=True,
                remove_last_layer=False,
                isize=isize,
                args=args,
            ),
        )

        # [h_img, h_topic] -> zd
        self.add_module(
            "encoder_cat_topic_img_h2zd",
            LSEncoderLinear(dim_input=self.img_h_dim + num_topics, z_dim=self.zd_dim),
        )

    def forward(self, img, vec_topic):
        """forward.

        :param x:
        :param topic:
        """
        # image->h_img
        h_img = self.layer_img2h4zd(img)
        h_topic = vec_topic
        h_img_topic = torch.cat((h_img, h_topic), 1)
        q_zd, zd_q = self.encoder_cat_topic_img_h2zd(h_img_topic)
        return q_zd, zd_q



import torch.nn as nn

from domainlab.compos.vae.compos.encoder import LSEncoderConvBnReluPool
from domainlab.compos.vae.compos.encoder_zy import EncoderConnectLastFeatLayer2Z
from domainlab.utils.utils_class import store_args


class XYDEncoderParallel(nn.Module):
    """
    calculate zx, zy, zd vars independently (without order, parallel):
    x->zx, x->zy, x->zd
    """

    def __init__(self, net_infer_zd, net_infer_zx, net_infer_zy):
        super().__init__()
        self.add_module("net_infer_zd", net_infer_zd)
        self.add_module("net_infer_zx", net_infer_zx)
        self.add_module("net_infer_zy", net_infer_zy)

    def forward(self, img):
        """
        encode img into 3 latent variables separately/parallel
        :param img:
        :return q_zd, zd_q, q_zx, zx_q, q_zy, zy_q
        """
        q_zd, zd_q = self.net_infer_zd(img)
        q_zx, zx_q = self.net_infer_zx(img)
        q_zy, zy_q = self.net_infer_zy(img)
        return q_zd, zd_q, q_zx, zx_q, q_zy, zy_q

    def infer_zy_loc(self, tensor):
        """
        Used by VAE model to predict class label
        """
        q_zy, _ = self.net_infer_zy(tensor)
        zy_loc = q_zy.mean
        return zy_loc


class XYDEncoderParallelUser(XYDEncoderParallel):
    """
    This class only reimplemented constructor of parent class
    """

    @store_args
    def __init__(self, net_class_d, net_x, net_class_y):
        super().__init__(net_class_d, net_x, net_class_y)


class XYDEncoderParallelConvBnReluPool(XYDEncoderParallel):
    """
    This class only reimplemented constructor of parent class
    """

    @store_args
    def __init__(self, zd_dim, zx_dim, zy_dim, i_c, i_h, i_w, conv_stride=1):
        """
        :param zd_dim:
        :param zx_dim:
        :param zy_dim:
        :param i_c: number of image channels
        :param i_h: image height
        :param i_w: image width
        """
        # conv_stride=2 on size 28 got RuntimeError:
        # Given input size: (64x1x1).
        # Calculated output size: (64x0x0).
        # Output size is too small
        net_infer_zd = LSEncoderConvBnReluPool(
            self.zd_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )
        # if self.zx_dim != 0:
        # pytorch can generate emtpy tensor, so no need to judge here
        net_infer_zx = LSEncoderConvBnReluPool(
            self.zx_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )
        net_infer_zy = LSEncoderConvBnReluPool(
            self.zy_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )
        super().__init__(net_infer_zd, net_infer_zx, net_infer_zy)


class XYDEncoderParallelAlex(XYDEncoderParallel):
    """
    This class only reimplemented constructor of parent class,
    at the end of the constructor of this class, the parent
    class contructor is called
    """

    @store_args
    def __init__(self, zd_dim, zx_dim, zy_dim, i_c, i_h, i_w, args, conv_stride=1):
        """
        :param zd_dim:
        :param zx_dim:
        :param zy_dim:
        :param i_c: number of image channels
        :param i_h: image height
        :param i_w: image width
        """
        # conv_stride=2 on size 28 got RuntimeError:
        # Given input size: (64x1x1).
        # Calculated output size: (64x0x0).
        # Output size is too small
        net_infer_zd = LSEncoderConvBnReluPool(
            self.zd_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )
        # if self.zx_dim != 0: pytorch can generate emtpy tensor,
        # so no need to judge here
        net_infer_zx = LSEncoderConvBnReluPool(
            self.zx_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )
        net_infer_zy = EncoderConnectLastFeatLayer2Z(
            self.zy_dim,
            True,
            i_c,
            i_h,
            i_w,
            args,
            arg_name="nname",
            arg_path_name="npath",
        )
        super().__init__(net_infer_zd, net_infer_zx, net_infer_zy)


class XYDEncoderParallelExtern(XYDEncoderParallel):
    """
    This class only reimplemented constructor of parent class,
    at the end of the constructor of this class, the parent
    class contructor is called
    """

    @store_args
    def __init__(self, zd_dim, zx_dim, zy_dim, args, i_c, i_h, i_w, conv_stride=1):
        """
        :param zd_dim:
        :param zx_dim:
        :param zy_dim:
        """
        net_infer_zd = EncoderConnectLastFeatLayer2Z(
            self.zd_dim,
            True,
            i_c,
            i_h,
            i_w,
            args,
            arg_name="nname_dom",
            arg_path_name="npath_dom",
        )
        # if self.zx_dim != 0: pytorch can generate emtpy tensor,
        # so no need to judge zx_dim=0 here
        net_infer_zx = LSEncoderConvBnReluPool(
            self.zx_dim, self.i_c, self.i_w, self.i_h, conv_stride=conv_stride
        )

        net_infer_zy = EncoderConnectLastFeatLayer2Z(
            self.zy_dim,
            True,
            i_c,
            i_h,
            i_w,
            args,
            arg_name="nname",
            arg_path_name="npath",
        )
        super().__init__(net_infer_zd, net_infer_zx, net_infer_zy)



"""
A logger for our software
"""
import logging
import multiprocessing
import os


class Logger:
    """
    static logger class
    """

    logger = None

    @staticmethod
    def get_logger(
        logger_name="logger_" + str(multiprocessing.current_process().pid),
        loglevel="INFO",
    ):
        """
        returns a logger
        if no logger was created yet, it will create a logger with the name
        specified in logger_name with the level specified in loglevel.
        If the logger was created for the first time the arguments do not change
        anything at the behaviour anymore
        """
        if Logger.logger is None:
            Logger.logger = logging.getLogger(logger_name)
            Logger.logger.setLevel(loglevel)

            # Create handlers and set their logging level
            logfolder = "zoutput/logs"
            os.makedirs(logfolder, exist_ok=True)
            # Create handlers and set their logging level
            filehandler = logging.FileHandler(
                logfolder + "/" + Logger.logger.name + ".log", mode="w"
            )
            filehandler.setLevel(loglevel)

            console_handler = logging.StreamHandler()
            console_handler.setLevel(loglevel)
            # Add handlers to logger
            Logger.logger.addHandler(filehandler)
            Logger.logger.addHandler(console_handler)
        return Logger.logger



"""
builder for erm
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.algos.utils import split_net_feat_last
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter
from domainlab.models.model_erm import mk_erm
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderERM(NodeAlgoBuilder):
    """
    builder for erm
    """

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        task = exp.task
        args = exp.args
        device = get_device(args)
        model_sel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
        observer = ObVisitor(model_sel)

        builder = FeatExtractNNBuilderChainNodeGetter(
            args, arg_name_of_net="nname", arg_path_of_net="npath"
        )()  # request, # @FIXME, constant string

        net = builder.init_business(
            flag_pretrain=True,
            dim_out=task.dim_y,
            remove_last_layer=False,
            args=args,
            isize=(task.isize.i_c, task.isize.i_h, task.isize.i_w),
        )

        _, _ = split_net_feat_last(net)

        model = mk_erm()(
            net=net,
            # net_feat=net_invar_feat, net_classifier=net_classifier,
            list_str_y=task.list_str_y,
        )
        model = self.init_next_model(model, exp)
        trainer = TrainerChainNodeGetter(args.trainer)(default="basic")
        # trainer.init_business(model, task, observer, device, args)
        return trainer, model, observer, device



"""
globals for the whole package
"""
__docformat__ = "restructuredtext"
import torch

g_inst_component_loss_agg = torch.sum
g_tensor_batch_agg = torch.sum
g_list_loss_agg = sum


def g_list_model_penalized_reg_agg(list_penalized_reg):
    """
    aggregate along the list, but do not diminish the batch structure of the tensor
    """
    return torch.stack(list_penalized_reg, dim=0).sum(dim=0)


g_str_cross_entropy_agg = "none"
# component loss refers to aggregation of pixel loss, digit of KL divergences loss
# instance loss currently use torch.sum, which is the same effect as torch.mean, the
# important part is the component aggregation method inside a single instance



"""
Task wraps around datasets, this file provide utilities
"""
import os
from pathlib import Path

import numpy
import torch
import torchvision
from torch.utils.data import Dataset

from domainlab.utils.utils_class import store_args


class ImSize:
    """ImSize."""

    @store_args
    def __init__(self, i_c, i_h, i_w):
        """
        store channel, height, width
        """

    @property
    def c(self):
        """image channel"""
        return self.i_c

    @property
    def h(self):
        """image height"""
        return self.i_h

    @property
    def w(self):
        """image width"""
        return self.i_w


def mk_onehot(dim, ind):
    """
    :param dim: dimension of representation vector
    :param ind: index
    """
    eye = torch.eye(dim)
    vec = eye[ind]
    return vec


def mk_loader(dset, bsize, drop_last=True, shuffle=True, num_workers=int(0)):
    """
    :param bs: batch size
    """
    if len(dset) < bsize:
        bsize = len(dset)
    loader = torch.utils.data.DataLoader(
        dataset=dset,
        batch_size=bsize,
        shuffle=shuffle,
        # @FIXME: shuffle must be true so the last incomplete batch get used in another epoch?
        num_workers=num_workers,  # @FIXME: num_workers=int(0) can be slow?
        drop_last=drop_last,
    )
    return loader


class DsetDomainVecDecorator(Dataset):
    """
    decorate a pytorch dataset with a fixed vector representation of domain
    """

    def __init__(self, dset, vec_domain, na_domain):
        """
        :param dset: x, y
        :param vec_domain: vector representation of domain
        :param na_domain: string description of domain
        """
        self.dset = dset
        self.vec_domain = vec_domain
        self.na_domain = na_domain

    @property
    def targets(self):
        """
        return a list of all targets so class sample count is straight forward
        """
        return self.dset.targets

    def __getitem__(self, idx):
        """
        :param idx:
        """
        tensor, vec_class, *other_vars = self.dset.__getitem__(idx)
        if other_vars:
            return (tensor, vec_class, self.vec_domain, *other_vars)
        return tensor, vec_class, self.vec_domain

    def __len__(self):
        """__len__."""
        return self.dset.__len__()


class DsetDomainVecDecoratorImgPath(DsetDomainVecDecorator):
    """
    Except returning x, y, d, additionally, the path of x is
    returned currently not in use since it is mostly important
    to print predictions together with path  for the test domain
    """

    def __getitem__(self, idx):
        """
        :param idx:
        """
        tensor, vec_class, path = self.dset.__getitem__(idx)
        return tensor, vec_class, self.vec_domain, path


class DsetClassVecDecorator(Dataset):
    """
    decorate a pytorch dataset with a new class name
    """

    def __init__(self, dset, dict_folder_name2class_global, list_str_y):
        """
        :param dset: x, y, *d
        :param dict_folder2class: dictionary that maps
                class folder of domain to glbal class
        """
        self.dset = dset
        self.class2idx = {
            k: v
            for (k, v) in self.dset.class_to_idx.items()
            if k in self.dset.list_class_dir
        }
        assert self.class2idx
        self.dict_folder_name2class_global = dict_folder_name2class_global
        self.list_str_y = list_str_y
        # inverst key:value to value:key for backward map
        self.dict_old_idx2old_class = dict((v, k) for k, v in self.class2idx.items())
        dict_class_na_local2vec_new = dict(
            (k, self.fun_class_local_na2vec_new(k)) for k, v in self.class2idx.items()
        )
        self.dict_class_na_local2vec_new = dict_class_na_local2vec_new

    @property
    def targets(self):
        """
        return a list of all targets so class sample count is straight forward
        """
        return self.dset.targets

    def fun_class_local_na2vec_new(self, k):
        """
        local class name within one domain, to one-hot
        vector of new representation
        """
        ind = self.list_str_y.index(self.dict_folder_name2class_global[k])
        return mk_onehot(len(self.list_str_y), ind)

    def __getitem__(self, idx):
        """
        :param idx:
        """
        tensor, vec_class, *other_vars = self.dset.__getitem__(idx)
        vec_class = vec_class.numpy()
        ind_old = numpy.argmax(vec_class)
        class_local = self.dict_old_idx2old_class[ind_old]
        vec_class_new = self.dict_class_na_local2vec_new[class_local]
        return tensor, vec_class_new, *other_vars

    def __len__(self):
        """__len__."""
        return self.dset.__len__()


class DsetClassVecDecoratorImgPath(DsetClassVecDecorator):
    def __getitem__(self, idx):
        """
        :param idx:
        This function is mainly
        """
        tensor, vec_class_new, path = super().__getitem__(idx)
        return tensor, vec_class_new, path[0]


class LoaderDomainLabel:
    """
    wraps a dataset with domain label and into a loader
    """

    def __init__(self, batch_size, dim_d):
        """__init__.

        :param batch_size:
        :param dim_d:
        """
        self.batch_size = batch_size
        self.dim_d = dim_d

    def __call__(self, dset, d_ind, na_domain):
        """
        wrap_dataset2loader_with_domain_label.
        :param dataset:
        :param batch_size:
        :param d_dim:
        :param d_ind:
        """
        d_eye = torch.eye(self.dim_d)
        d_label = d_eye[d_ind]
        dset = DsetDomainVecDecorator(dset, d_label, na_domain)
        loader = mk_loader(dset, self.batch_size)
        return loader


def tensor1hot2ind(tensor_label):
    """tensor1hot2ind.

    :param tensor_label:
    """
    _, label_ind = torch.max(tensor_label, dim=1)
    npa_label_ind = label_ind.numpy()
    return npa_label_ind


# @FIXME: this function couples strongly with the task,
# should be a class method of task
def img_loader2dir(
    loader, folder, test=False, list_domain_na=None, list_class_na=None, batches=5
):
    """
    save images from loader to directory so speculate if loader is correct
    :param loader: pytorch data loader
    :param folder: folder to save images
    :param test: if true, the loader is assumend to be a test loader; if false (default) it is assumed to be a train loader
    :param list_domain_na: optional list of domain names
    :param list_class_na: optional list of class names
    :param batches: number of batches to save
    """
    Path(os.path.normpath(folder)).mkdir(parents=True, exist_ok=True)
    l_iter = iter(loader)
    counter = 0
    batches = min(batches, len(l_iter))
    for _ in range(batches):
        img, vec_y, *other_vars = next(l_iter)
        class_label_ind_batch = tensor1hot2ind(vec_y)

        # get domain label
        # Note 1: test loaders don't return domain labels (see NodeTaskDictClassif.init_business)
        # Note 2: for train loaders domain label will be the 0th element of other_vars (see DsetDomainVecDecorator class above)
        has_domain_label_ind = False
        if not test:
            if other_vars:
                domain_label_ind_batch = tensor1hot2ind(other_vars[0])
                has_domain_label_ind = True

        for b_ind in range(img.shape[0]):
            class_label_ind = class_label_ind_batch[b_ind]
            class_label_scalar = class_label_ind.item()

            if list_class_na is None:
                str_class_label = "class_" + str(class_label_scalar)
            else:
                # @FIXME: where is the correspndance between
                # class ind_label and class str_label?
                str_class_label = list_class_na[class_label_scalar]
            str_domain_label = "unknown"
            if has_domain_label_ind:
                domain_label_ind = domain_label_ind_batch[b_ind]
                if list_domain_na is None:
                    str_domain_label = str(domain_label_ind)
                else:
                    # @FIXME: the correspondance between
                    # domain ind_label and domain str_label is missing
                    str_domain_label = list_domain_na[domain_label_ind]
            arr = img[b_ind]
            img_vision = torchvision.transforms.ToPILImage()(arr)
            f_n = "_".join(
                [
                    "class",
                    str_class_label,
                    "domain",
                    str_domain_label,
                    "n",
                    str(counter),
                ]
            )
            counter += 1
            path = os.path.join(folder, f_n + ".png")
            img_vision.save(path)



"""
network builder utils
"""
from torch import nn


def split_net_feat_last(net):
    """
    split a pytorch module into feature extractor and last layer
    """
    net_classifier = list(net.children())[-1]
    net_invar_feat = nn.Sequential(*(list(net.children())[:-1]))
    return net_invar_feat, net_classifier



"""
task specific dataset operation
"""
import random

from torch.utils.data import Dataset


class DsetIndDecorator4XYD(Dataset):
    """
    For dataset of x, y, d,  decorate it wih index
    """

    def __init__(self, dset):
        """
        :param dset: x,y,d
        """
        tuple_m = dset[0]
        if len(tuple_m) < 3:
            raise RuntimeError(
                "dataset to be wrapped should output at least x, y, and d; got length ",
                len(tuple_m),
            )
        self.dset = dset

    def __getitem__(self, index):
        """
        :param index:
        """
        tensor_x, vec_y, vec_d, *_ = self.dset.__getitem__(index)
        return tensor_x, vec_y, vec_d, index

    def __len__(self):
        return self.dset.__len__()


class DsetZip(Dataset):
    """
    enable zip return in getitem: x_1, y_1, x_2, y_2
    to avoid always the same match, the second dataset does not use the same idx in __get__item()
    but instead, a random one
    """

    def __init__(self, dset1, dset2, name=None):
        """
        :param dset1: x1, y1, *d1
        :param dset2: x2, y2, *d2
        :param name: name of dataset
        """
        self.dset1 = dset1
        self.dset2 = dset2
        self.name = name
        self.len2 = self.dset2.__len__()

    def __getitem__(self, idx):
        """
        :param idx:
        """
        idx2 = idx + random.randrange(self.len2)
        idx2 = idx2 % self.len2
        tensor_x_1, vec_y_1, vec_d_1, *others_1 = self.dset1.__getitem__(idx)
        tensor_x_2, vec_y_2, vec_d_2, *others_2 = self.dset2.__getitem__(idx2)
        return (
            tensor_x_1,
            vec_y_1,
            vec_d_1,
            others_1,
            tensor_x_2,
            vec_y_2,
            vec_d_2,
            others_2,
        )

    def __len__(self):
        len1 = self.dset1.__len__()
        if len1 < self.len2:
            return len1
        return self.len2



"""
builder for JiGen
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.observers.c_obvisitor_cleanup import ObVisitorCleanUp
from domainlab.algos.trainers.hyper_scheduler import HyperSchedulerWarmupExponential
from domainlab.algos.trainers.train_hyper_scheduler import TrainerHyperScheduler
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.compos.nn_zoo.net_classif import ClassifDropoutReluLinear
from domainlab.compos.utils_conv_get_flat_dim import get_flat_dim
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter
from domainlab.dsets.utils_wrapdset_patches import WrapDsetPatches
from domainlab.models.model_jigen import mk_jigen
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderJiGen(NodeAlgoBuilder):
    """
    NodeAlgoBuilderJiGen
    """

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        task = exp.task
        args = exp.args
        device = get_device(args)
        msel = MSelOracleVisitor(msel=MSelValPerf(max_es=args.es))
        observer = ObVisitor(msel)
        observer = ObVisitorCleanUp(observer)

        builder = FeatExtractNNBuilderChainNodeGetter(
            args, arg_name_of_net="nname", arg_path_of_net="npath"
        )()  # request, @FIXME, constant string

        net_encoder = builder.init_business(
            flag_pretrain=True,
            dim_out=task.dim_y,
            remove_last_layer=False,
            args=args,
            isize=(task.isize.i_c, task.isize.i_w, task.isize.i_h),
        )

        dim_feat = get_flat_dim(
            net_encoder, task.isize.i_c, task.isize.i_h, task.isize.i_w
        )

        net_classifier = ClassifDropoutReluLinear(dim_feat, task.dim_y)

        # @FIXME: this seems to be the only difference w.r.t. builder_dann
        net_classifier_perm = ClassifDropoutReluLinear(dim_feat, args.nperm + 1)
        model = mk_jigen()(
            list_str_y=task.list_str_y,
            coeff_reg=args.gamma_reg,
            net_encoder=net_encoder,
            net_classifier_class=net_classifier,
            net_classifier_permutation=net_classifier_perm,
            n_perm=args.nperm,
            prob_permutation=args.pperm,
        )
        model = self.init_next_model(model, exp)
        trainer = TrainerChainNodeGetter(args.trainer)(default="hyperscheduler")
        trainer.init_business(model, task, observer, device, args)
        if isinstance(trainer, TrainerHyperScheduler):
            trainer.set_scheduler(
                HyperSchedulerWarmupExponential,
                total_steps=trainer.num_batches * args.warmup,
                flag_update_epoch=False,
                flag_update_batch=True,
            )
        return trainer, model, observer, device



"""
Command line arguments
"""
import argparse
import warnings

import yaml

from domainlab.algos.trainers.args_dial import add_args2parser_dial
from domainlab.algos.trainers.compos.matchdg_args import add_args2parser_matchdg
from domainlab.models.args_jigen import add_args2parser_jigen
from domainlab.models.args_vae import add_args2parser_vae
from domainlab.utils.logger import Logger


def mk_parser_main():
    """
    Args for command line definition
    """
    parser = argparse.ArgumentParser(description="DomainLab")

    parser.add_argument(
        "-c",
        "--config",
        default=None,
        help="load YAML configuration",
        dest="config_file",
        type=argparse.FileType(mode="r"),
    )

    parser.add_argument("--lr", type=float, default=1e-4, help="learning rate")

    parser.add_argument(
        "--gamma_reg", type=float, default=0.1, help="weight of regularization loss"
    )

    parser.add_argument("--es", type=int, default=1, help="early stop steps")

    parser.add_argument("--seed", type=int, default=0, help="random seed (default: 0)")

    parser.add_argument(
        "--nocu", action="store_true", default=False, help="disables CUDA"
    )

    parser.add_argument(
        "--device", type=str, default=None, help="device name default None"
    )

    parser.add_argument(
        "--gen", action="store_true", default=False, help="save generated images"
    )

    parser.add_argument(
        "--keep_model",
        action="store_true",
        default=False,
        help="do not delete model at the end of training",
    )

    parser.add_argument("--epos", default=2, type=int, help="maximum number of epochs")

    parser.add_argument(
        "--epos_min", default=0, type=int, help="maximum number of epochs"
    )

    parser.add_argument(
        "--epo_te", default=1, type=int, help="test performance per {} epochs"
    )

    parser.add_argument(
        "-w",
        "--warmup",
        type=int,
        default=100,
        help="number of epochs for hyper-parameter warm-up. \
                        Set to 0 to turn warmup off.",
    )

    parser.add_argument("--debug", action="store_true", default=False)
    parser.add_argument("--dmem", action="store_true", default=False)
    parser.add_argument(
        "--no_dump",
        action="store_true",
        default=False,
        help="suppress saving the confusion matrix",
    )

    parser.add_argument(
        "--trainer", type=str, default=None, help="specify which trainer to use"
    )

    parser.add_argument(
        "--out", type=str, default="zoutput", help="absolute directory to store outputs"
    )

    parser.add_argument(
        "--dpath",
        type=str,
        default="zdpath",
        help="path for storing downloaded dataset",
    )

    parser.add_argument(
        "--tpath",
        type=str,
        default=None,
        help="path for custom task, should implement \
                        get_task function",
    )

    parser.add_argument(
        "--npath",
        type=str,
        default=None,
        help="path of custom neural network for feature \
                        extraction",
    )

    parser.add_argument(
        "--npath_dom",
        type=str,
        default=None,
        help="path of custom neural network for feature \
                        extraction",
    )

    parser.add_argument(
        "--npath_argna2val",
        action="append",
        help="specify new arguments and their value \
                        e.g. '--npath_argna2val my_custom_arg_na \
                        --npath_argna2val xx/yy/zz.py', additional \
                        pairs can be appended",
    )

    parser.add_argument(
        "--nname_argna2val",
        action="append",
        help="specify new arguments and their values \
                        e.g. '--nname_argna2val my_custom_network_arg_na \
                        --nname_argna2val alexnet', additional pairs \
                        can be appended",
    )

    parser.add_argument(
        "--nname",
        type=str,
        default=None,
        help="name of custom neural network for feature \
                        extraction of classification",
    )

    parser.add_argument(
        "--nname_dom",
        type=str,
        default=None,
        help="name of custom neural network for feature \
                        extraction of domain",
    )

    parser.add_argument(
        "--apath", type=str, default=None, help="path for custom AlgorithmBuilder"
    )

    parser.add_argument(
        "--exptag",
        type=str,
        default="exptag",
        help="tag as prefix of result aggregation file name \
                        e.g. git hash for reproducibility",
    )

    parser.add_argument(
        "--aggtag",
        type=str,
        default="aggtag",
        help="tag in each line of result aggregation file \
                        e.g., to specify potential different configurations",
    )

    parser.add_argument(
        "--agg_partial_bm",
        type=str,
        default=None,
        dest="bm_dir",
        help="Aggregates and plots partial data of a snakemake \
                        benchmark. Requires the benchmark config file. \
                        Other arguments will be ignored.",
    )

    parser.add_argument(
        "--gen_plots",
        type=str,
        default=None,
        dest="plot_data",
        help="plots the data of a snakemake benchmark. "
        "Requires the results.csv file"
        "and an output file (specify by --outp_file,"
        "default is zoutput/benchmarks/shell_benchmark). "
        "Other arguments will be ignored.",
    )

    parser.add_argument(
        "--outp_dir",
        type=str,
        default="zoutput/benchmarks/shell_benchmark",
        dest="outp_dir",
        help="outpus file for the plots when creating them"
        "using --gen_plots. "
        "Default is zoutput/benchmarks/shell_benchmark",
    )

    parser.add_argument(
        "--param_idx",
        type=bool,
        default=True,
        dest="param_idx",
        help="True: parameter index is used in the "
        "pots generated with --gen_plots."
        "False: parameter name is used."
        "Default is True.",
    )

    parser.add_argument(
        "--msel",
        choices=["val", "loss_tr"],
        default="val",
        help="model selection for early stop: val, loss_tr, recon, the \
                        elbo and recon only make sense for vae models,\
                        will be ignored by other methods",
    )

    parser.add_argument(
        "--model", metavar="an", type=str, default=None, help="algorithm name"
    )

    parser.add_argument(
        "--acon",
        metavar="ac",
        type=str,
        default=None,
        help="algorithm configuration name, (default None)",
    )

    parser.add_argument("--task", metavar="ta", type=str, help="task name")

    arg_group_task = parser.add_argument_group("task args")

    arg_group_task.add_argument(
        "--bs", type=int, default=100, help="loader batch size for mixed domains"
    )

    arg_group_task.add_argument(
        "--split",
        type=float,
        default=0,
        help="proportion of training, a value between \
                                0 and 1, 0 means no train-validation split",
    )

    arg_group_task.add_argument(
        "--te_d",
        nargs="*",
        default=None,
        help="test domain names separated by single space, \
                                will be parsed to be list of strings",
    )

    arg_group_task.add_argument(
        "--tr_d",
        nargs="*",
        default=None,
        help="training domain names separated by \
                                single space, will be parsed to be list of \
                                strings; if not provided then all available \
                                domains that are not assigned to \
                                the test set will be used as training domains",
    )

    arg_group_task.add_argument(
        "--san_check",
        action="store_true",
        default=False,
        help="save images from the dataset as a sanity check",
    )

    arg_group_task.add_argument(
        "--san_num",
        type=int,
        default=8,
        help="number of images to be dumped for the sanity check",
    )

    arg_group_task.add_argument(
        "--loglevel", type=str, default="DEBUG", help="sets the loglevel of the logger"
    )

    # args for variational auto encoder
    arg_group_vae = parser.add_argument_group("vae")
    arg_group_vae = add_args2parser_vae(arg_group_vae)
    arg_group_matchdg = parser.add_argument_group("matchdg")
    arg_group_matchdg = add_args2parser_matchdg(arg_group_matchdg)
    arg_group_jigen = parser.add_argument_group("jigen")
    arg_group_jigen = add_args2parser_jigen(arg_group_jigen)
    args_group_dial = parser.add_argument_group("dial")
    args_group_dial = add_args2parser_dial(args_group_dial)
    return parser


def apply_dict_to_args(args, data: dict, extend=False):
    """
    Tries to apply the data to the args dict of DomainLab.
    Unknown keys are silently ignored as long as
    extend is not set.
    """
    arg_dict = args.__dict__
    for key, value in data.items():
        if (key in arg_dict) or extend:
            if isinstance(value, list):
                cur_val = arg_dict.get(key, None)
                if not isinstance(cur_val, list):
                    if cur_val is not None:
                        raise RuntimeError(
                            f"input dictionary value is list, \
                                           however, in DomainLab args, we have {cur_val}, \
                                           going to overrite to list"
                        )
                    arg_dict[key] = []  # if args_dict[key] is None, cast it into a list
                    # domainlab will take care of it if this argument can not be a list
                arg_dict[key].extend(value)  # args_dict[key] is already a list
                # keep existing values for the list arg_dct[key]
            else:
                # over-write existing value
                arg_dict[key] = value
        else:
            raise ValueError("Unsupported key: ", key)


def parse_cmd_args():
    """
    get args from command line
    """
    parser = mk_parser_main()
    args = parser.parse_args()
    logger = Logger.get_logger(logger_name="main_out_logger", loglevel=args.loglevel)
    if args.config_file:
        data = yaml.safe_load(args.config_file)
        delattr(args, "config_file")
        apply_dict_to_args(args, data)

    if args.acon is None and args.bm_dir is None:
        logger.warn("\n\n")
        logger.warn("no algorithm conf specified, going to use default")
        logger.warn("\n\n")
        warnings.warn("no algorithm conf specified, going to use default")

    return args



"""
When class names and numbers does not match across different domains
"""
from domainlab.tasks.task_folder import NodeTaskFolderClassNaMismatch


def mk_task_folder(
    extensions,
    list_str_y,
    dict_domain_folder_name2class,
    dict_domain_img_trans,
    img_trans_te,
    isize,
    dict_domain2imgroot,
    taskna,
    succ=None,
):
    """
    Make task by specifying each domain with folder structures
    :param extensions: Different Options: 1. a python dictionary with key as the domain name
    and value (str or tuple[str]) as the file extensions of the image. 2. a str or tuple[str]
    with file extensions for all domains. 3. None: in each domain all files with an extension
    in ('jpg', 'jpeg', 'png') are loaded.
    :param
    list_str_y: a python list with user defined class name where
    the order of the list matters.
    :param dict_domain_folder_name2class: a python dictionary, with key
    as the user specified domain name, value as a dictionary to map the
    sub-folder name of each domain's class folder into the user defined
    common class name.
    :param dict_domain_img_trans: a python dictionary with keys as the user
    specified domain name, value as a user defined torchvision transform.
    This feature allows carrying out different transformation (composition) to different
    domains at training time.
    :param img_trans_te: at test or inference time, we do not have knowledge
    of domain information so only a unique transform (composition) is allowed.
    :isize: domainlab.tasks.ImSize(image channel, image height, image width)
    :dict_domain2imgroot: a python dictionary with keys as user specified domain
    names and values as the absolute path to each domain's data.
    :taskna: user defined task name
    """

    class NodeTaskFolderDummy(NodeTaskFolderClassNaMismatch):
        @property
        def task_name(self):
            """
            The basic name of the task, without configurations
            """
            return taskna

        def conf_without_args(self):
            self.extensions = extensions
            self.list_str_y = list_str_y
            self._dict_domain_folder_name2class = dict_domain_folder_name2class
            self.isize = isize
            self.set_list_domains(list(self._dict_domain_folder_name2class.keys()))
            self.dict_domain2imgroot = dict_domain2imgroot
            self._dict_domain_img_trans = dict_domain_img_trans
            self.img_trans_te = img_trans_te

    return NodeTaskFolderDummy(succ=succ)



import os

import torch

from domainlab.compos.vae.c_vae_recon import ReconVAEXYD
from domainlab.utils.utils_img_sav import mk_fun_sav_img


class FlowGenImgs:
    def __init__(self, model, device):
        model = model.to(device)
        self.obj_recon = ReconVAEXYD(model)
        self.sav_fun = None

    def _save_list(self, recon_list_d, device, img, name):
        """
        compare counterfactual
        """
        recon_stack_d = torch.cat(recon_list_d)
        recon_stack_d = recon_stack_d.to(device)
        comparison = torch.cat([img, recon_stack_d])
        self.sav_fun(comparison, name, title=None)  # row is actually column here

    def _save_pair(self, x_recon_img, device, img, name):
        """
        compare recon and original
        """
        x_recon_img = x_recon_img.to(device)
        comparison = torch.cat([img, x_recon_img])
        self.sav_fun(comparison, name)

    def gen_img_loader(self, loader, device, path, domain):
        """
        gen image for the first batch of input loader
        """
        for _, (x_batch, y_batch, *d_batch) in enumerate(loader):
            x_batch, y_batch = x_batch.to(device), y_batch.to(device)
            if d_batch and isinstance(d_batch[0], torch.Tensor):
                d_batch = d_batch[0].to(device)
            else:
                d_batch = None
            self.gen_img_xyd(x_batch, y_batch, d_batch, device, path, folder_na=domain)
            break

    def gen_img_xyd(self, img, vec_y, vec_d, device, path, folder_na):
        img = img.to(device)
        vec_y = vec_y.to(device)
        if vec_d is not None:
            vec_d = vec_d.to(device)
        nrow = img.shape[0]  # nrow is actually ncol in pytorch save_image!!
        self.sav_fun = mk_fun_sav_img(path=path, nrow=nrow, folder_na=folder_na)
        self._flow_vanilla(img, vec_y, vec_d, device)
        self._flow_cf_y(img, vec_y, vec_d, device)
        if vec_d is not None:
            self._flow_cf_d(img, vec_y, vec_d, device)

    def _flow_vanilla(self, img, vec_y, vec_d, device, num_sample=10):
        x_recon_img, str_type = self.obj_recon.recon(img)
        self._save_pair(x_recon_img, device, img, str_type + ".png")

        x_recon_img, str_type = self.obj_recon.recon(img, vec_y)
        self._save_pair(x_recon_img, device, img, str_type + ".png")

        if vec_d is not None:
            x_recon_img, str_type = self.obj_recon.recon(img, None, vec_d)
            self._save_pair(x_recon_img, device, img, str_type + ".png")

        for i in range(num_sample):
            x_recon_img, str_type = self.obj_recon.recon(img, vec_y, vec_d, True, True)
            x_recon_img = x_recon_img.to(device)
            comparison = torch.cat([img, x_recon_img])
            self.sav_fun(comparison, str_type + str(i) + ".png")

    def _flow_cf_y(self, img, vec_y, vec_d, device):
        """
        scan possible values of vec_y
        """
        recon_list, str_type = self.obj_recon.recon_cf(
            img, "y", vec_y.shape[1], device, zx2fill=None
        )
        self._save_list(
            recon_list, device, img, "_".join(["recon_cf_y", str_type]) + ".png"
        )
        recon_list, str_type = self.obj_recon.recon_cf(
            img, "y", vec_y.shape[1], device, zx2fill=0
        )
        self._save_list(
            recon_list, device, img, "_".join(["recon_cf_y", str_type]) + ".png"
        )
        if vec_d is not None:
            recon_list, str_type = self.obj_recon.recon_cf(
                img, "y", vec_y.shape[1], device, vec_d=vec_d, zx2fill=0
            )
            self._save_list(
                recon_list, device, img, "_".join(["recon_cf_y", str_type]) + ".png"
            )

    def _flow_cf_d(self, img, vec_y, vec_d, device):
        """
        scan possible values of vec_y
        """
        recon_list, str_type = self.obj_recon.recon_cf(
            img, "d", vec_d.shape[1], device, zx2fill=None
        )
        self._save_list(
            recon_list, device, img, "_".join(["recon_cf_d", str_type]) + ".png"
        )
        recon_list, str_type = self.obj_recon.recon_cf(
            img, "d", vec_d.shape[1], device, zx2fill=0
        )
        self._save_list(
            recon_list, device, img, "_".join(["recon_cf_d", str_type]) + ".png"
        )


def fun_gen(model, device, node, args, subfolder_na, output_folder_na="gen"):
    flow = FlowGenImgs(model, device)
    path = os.path.join(
        args.out, output_folder_na, node.task_name, args.model, subfolder_na
    )
    flow.gen_img_loader(node.loader_te, device, path=path, domain="_".join(args.te_d))
    flow.gen_img_loader(
        node.loader_tr, device, path=path, domain="_".join(node.list_domain_tr)
    )



"""
import external neural network implementation
"""

from domainlab.utils.logger import Logger
from domainlab.utils.u_import import import_path


def build_external_obj_net_module_feat_extract(mpath, dim_y, remove_last_layer):
    """The user provide a function to initiate an object of the neural network,
    which is fine for training but problematic for persistence of the trained
    model since it is created externally.
    :param mpath: path of external python file where the neural network
    architecture is defined
    :param dim_y: dimension of features
    """
    # other possibility
    # na_external_module = "name_external_module"   # the dummy module name
    # spec = importlib.util.spec_from_file_location(
    #    name=na_external_module,
    #    location=path_net_feat_extract)
    # module_external = importlib.util.module_from_spec(spec)
    # sys.modules[na_external_module] = module_external
    # register the name of the external module
    # spec.loader.exec_module(module_external)

    net_module = import_path(mpath)
    name_signature = "build_feat_extract_net(dim_y, \
        remove_last_layer)"
    # @FIXME: hard coded, move to top level __init__ definition in domainlab
    name_fun = name_signature[: name_signature.index("(")]
    if hasattr(net_module, name_fun):
        try:
            net = getattr(net_module, name_fun)(dim_y, remove_last_layer)
        except Exception:
            logger = Logger.get_logger()
            logger.error(
                f"function {name_signature} should return a neural network "
                f"(pytorch module) that that extract features from an image"
            )
            raise
        if net is None:
            raise RuntimeError(
                "the pytorch module returned by %s is None" % (name_signature)
            )
    else:
        raise RuntimeError(
            "Please implement a function %s \
                            in your external python file"
            % (name_signature)
        )
    return net



"""
Base class for Task
"""
import warnings
from abc import abstractmethod

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler
from domainlab.tasks.task_utils import parse_domain_id
from domainlab.utils.logger import Logger


class NodeTaskDG(AbstractChainNodeHandler):
    """
    Domain Generalization Classification Task
    """

    def __init__(self, succ=None):
        super().__init__(succ)
        self._loader_tr = None
        self._loader_te = None
        self._loader_val = None
        self._list_domains = None
        self._list_domain_tr = None  # versatile
        self._name = None
        self._args = None
        self.dict_dset_all = {}  # persist
        self.dict_dset_tr = {}  # versatile variable: which domains to use as training
        self.dict_dset_te = {}  # versatile
        self.dict_dset_val = {}  # versatile
        self.dict_domain_class_count = {}
        self.dim_d_tr = None  # public, only used for diva
        self._im_size = None
        self._dict_domains2imgroot = {}
        self._dict_domain_folder_name2class = (
            {}
        )  # {"domain1": {"class1":car, "class2":dog}}
        self._dict_domain_img_trans = {}
        self.dict_att = {}
        self.img_trans_te = None
        self.dict_domain2imgroot = {}
        self._dict_domain2filepath_list_im_tr = (
            {}
        )  # {"photo": "xxx/yyy/file_of_path2imgs"}
        self._dict_domain2filepath_list_im_val = {}
        self._dict_domain2filepath_list_im_te = {}
        self.dict_class_label_ind2name = None
        self.conf_without_args()  # configuration without init_business

    def conf_without_args(self):
        """
        configuration without init_business
        """

    @abstractmethod
    def init_business(self, args, trainer=None):
        """
        construct task data loader
        """

    def get_list_domains(self):
        """
        1. get list of domain names
        2. better use method than property so new domains can be added
        """
        return self._list_domains

    def set_list_domains(self, list_domains):
        """
        setter for self._list_domains
        """
        self._list_domains = list_domains

    @property
    def isize(self):
        """
        getter for input size: isize
        """
        return self._im_size

    @isize.setter
    def isize(self, im_size):
        """
        setter for input size: isize
        """
        self._im_size = im_size

    @property
    def list_domain_tr(self):
        """
        property getter of list of domains for this task
        """
        if self._list_domain_tr is None:
            raise RuntimeError("task not intialized!")
        return self._list_domain_tr

    @property
    def loader_tr(self):
        """loader of mixed train domains"""
        return self._loader_tr

    @property
    def loader_val(self):
        """loader of validation dataset on the training domains"""
        return self._loader_val

    @property
    def loader_te(self):
        """loader of mixed test domains"""
        return self._loader_te

    @property
    def task_name(self):
        """
        The basic name of the task, without configurations
        """
        # @FIXME: hardcoded position:NodeTaskXXX
        return type(self).__name__[8:].lower()

    def get_na(self, na_tr, na_te):
        """
        task name appended with configurations
        :param na_tr: training domain names
        :param na_te: test domain names
        """
        _, list_te = self.get_list_domains_tr_te(na_tr, na_te)
        str_te = "_".join(list_te)
        # train domain names are too long
        return "_".join([self.task_name, "te", str_te])

    def is_myjob(self, request):
        """
        :param request: string
        """
        return request == self.task_name

    def get_list_domains_tr_te(self, tr_id, te_id):
        """
        For static DG task, get train and test domains list.

        :param tr_id: training domain ids;
            int or str, or a list of int or str, or None;
            if None, then assumed to be the complement of te_id.
        :param te_id: test domain ids;
            int or str, or a list of int or str; required.
        :return: list of training domain names, list of test domain names.
        """
        list_domains = self.get_list_domains()

        list_domain_te = parse_domain_id(te_id, list_domains)
        assert set(list_domain_te).issubset(set(list_domains))

        if tr_id is None:
            list_domain_tr = [did for did in list_domains if did not in list_domain_te]
        else:
            list_domain_tr = parse_domain_id(tr_id, list_domains)
        if not set(list_domain_tr).issubset(set(list_domains)):
            raise RuntimeError(
                f"training domain {list_domain_tr} is not \
                subset of available domains {list_domains}"
            )

        if set(list_domain_tr) & set(list_domain_te):
            logger = Logger.get_logger()
            logger.warn(
                "The sets of training and test domains overlap -- "
                "be aware of data leakage or training to the test!"
            )
            warnings.warn(
                "The sets of training and test domains overlap -- "
                "be aware of data leakage or training to the test!",
                RuntimeWarning,
            )

        self.dim_d_tr = len(list_domain_tr)
        self._list_domain_tr = list_domain_tr
        return list_domain_tr, list_domain_te

    def __str__(self):
        """
        print the attribute of the task
        """
        strout = "list of domains: \n"
        strout += str(self.get_list_domains())
        strout += f"\n input tensor size: {self.isize}"
        return strout



"""
builder for Domain Adversarial Neural Network: accept different training scheme
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.observers.c_obvisitor_cleanup import ObVisitorCleanUp
from domainlab.algos.trainers.hyper_scheduler import HyperSchedulerWarmupExponential
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.compos.nn_zoo.net_classif import ClassifDropoutReluLinear
from domainlab.compos.utils_conv_get_flat_dim import get_flat_dim
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter
from domainlab.models.model_dann import mk_dann
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderDANN(NodeAlgoBuilder):
    """
    NodeAlgoBuilderDANN
    """

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        task = exp.task
        self._task = task
        args = exp.args
        task.get_list_domains_tr_te(args.tr_d, args.te_d)
        device = get_device(args)
        msel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
        observer = ObVisitor(msel)
        observer = ObVisitorCleanUp(observer)

        builder = FeatExtractNNBuilderChainNodeGetter(
            args, arg_name_of_net="nname", arg_path_of_net="npath"
        )()  # request, @FIXME, constant string

        net_encoder = builder.init_business(
            flag_pretrain=True,
            dim_out=task.dim_y,
            remove_last_layer=False,
            args=args,
            isize=(task.isize.i_c, task.isize.i_w, task.isize.i_h),
        )

        dim_feat = get_flat_dim(
            net_encoder, task.isize.i_c, task.isize.i_h, task.isize.i_w
        )

        net_classifier = ClassifDropoutReluLinear(dim_feat, task.dim_y)

        net_discriminator = self.reset_aux_net(net_encoder)
        model = mk_dann()(
            list_str_y=task.list_str_y,
            list_d_tr=task.list_domain_tr,
            alpha=args.gamma_reg,
            net_encoder=net_encoder,
            net_classifier=net_classifier,
            net_discriminator=net_discriminator,
            builder=self,
        )

        model = self.init_next_model(model, exp)
        trainer = TrainerChainNodeGetter(args.trainer)(default="hyperscheduler")
        trainer.init_business(model, task, observer, device, args)
        if trainer.name == "hyperscheduler":
            trainer.set_scheduler(
                HyperSchedulerWarmupExponential,
                total_steps=trainer.num_batches * args.warmup,
                flag_update_epoch=False,
                flag_update_batch=True,
            )
        return trainer, model, observer, device

    def reset_aux_net(self, net_encoder):
        """
        reset auxilliary neural network from task
        note that net_encoder can also be a method like extract_semantic_feat
        """
        dim_feat = get_flat_dim(
            net_encoder,
            self._task.isize.i_c,
            self._task.isize.i_h,
            self._task.isize.i_w,
        )
        net_discriminator = ClassifDropoutReluLinear(
            dim_feat, len(self._task.list_domain_tr)
        )
        return net_discriminator



"""
Meta Learning Domain Generalization
"""
import copy
import random

from torch.utils.data.dataset import ConcatDataset

from domainlab.algos.trainers.a_trainer import AbstractTrainer
from domainlab.algos.trainers.train_basic import TrainerBasic
from domainlab.tasks.utils_task import mk_loader
from domainlab.tasks.utils_task_dset import DsetZip


class TrainerMLDG(AbstractTrainer):
    """
    basic trainer
    """

    def before_tr(self):
        """
        check the performance of randomly initialized weight
        """
        self.model.evaluate(self.loader_te, self.device)
        self.inner_trainer = TrainerBasic()
        self.inner_trainer.extend(self._decoratee)
        inner_model = copy.deepcopy(self.model)
        self.inner_trainer.init_business(
            inner_model,
            copy.deepcopy(self.task),
            self.observer,
            self.device,
            self.aconf,
            flag_accept=False,
        )
        self.prepare_ziped_loader()

    def prepare_ziped_loader(self):
        """
        create virtual source and target domain
        """
        list_dsets = list(self.task.dict_dset_tr.values())
        num_domains = len(list_dsets)
        ind_target_domain = random.randrange(num_domains)
        tuple_dsets_source = tuple(
            list_dsets[ind] for ind in range(num_domains) if ind != ind_target_domain
        )
        ddset_source = ConcatDataset(tuple_dsets_source)
        ddset_target = list_dsets[ind_target_domain]
        ddset_mix = DsetZip(ddset_source, ddset_target)
        self.loader_tr_source_target = mk_loader(ddset_mix, self.aconf.bs)

    def tr_epoch(self, epoch):
        self.model.train()
        self.epo_loss_tr = 0
        self.prepare_ziped_loader()
        # s means source, t means target
        for ind_batch, (
            tensor_x_s,
            vec_y_s,
            vec_d_s,
            others_s,
            tensor_x_t,
            vec_y_t,
            vec_d_t,
            *_,
        ) in enumerate(self.loader_tr_source_target):
            tensor_x_s, vec_y_s, vec_d_s = (
                tensor_x_s.to(self.device),
                vec_y_s.to(self.device),
                vec_d_s.to(self.device),
            )

            tensor_x_t, vec_y_t, vec_d_t = (
                tensor_x_t.to(self.device),
                vec_y_t.to(self.device),
                vec_d_t.to(self.device),
            )

            self.optimizer.zero_grad()

            self.inner_trainer.model.load_state_dict(self.model.state_dict())
            # update inner_model
            self.inner_trainer.before_epoch()  # set model to train mode
            self.inner_trainer.reset()  # force optimizer to re-initialize
            self.inner_trainer.tr_batch(
                tensor_x_s, vec_y_s, vec_d_s, others_s, ind_batch, epoch
            )
            # inner_model has now accumulated gradients Gi
            # with parameters theta_i - lr * G_i where i index batch

            loss_look_forward = self.inner_trainer.model.cal_task_loss(
                tensor_x_t, vec_y_t
            )
            loss_source_task = self.model.cal_task_loss(tensor_x_s, vec_y_s)
            list_source_reg_tr, list_source_mu_tr = self.cal_reg_loss(
                tensor_x_s, vec_y_s, vec_d_s, others_s
            )
            # call cal_reg_loss from decoratee
            # super()._cal_reg_loss returns [],[],
            # since mldg's reg loss is on target domain,
            # no other trainer except hyperscheduler could decorate it unless we use state pattern
            # in the future to control source and target domain loader behavior
            source_reg_tr = self.model.list_inner_product(
                list_source_reg_tr, list_source_mu_tr
            )
            # self.aconf.gamma_reg * loss_look_forward.sum()
            loss = (
                loss_source_task.sum()
                + source_reg_tr.sum()
                + self.aconf.gamma_reg * loss_look_forward.sum()
            )
            #
            loss.backward()
            # optimizer only optimize parameters of self.model, not inner_model
            self.optimizer.step()
            self.epo_loss_tr += loss.detach().item()
            self.after_batch(epoch, ind_batch)
        flag_stop = self.observer.update(epoch)  # notify observer
        return flag_stop



"""
command-line interface (CLI) for the domainlab package
"""

from domainlab.arg_parser import parse_cmd_args
from domainlab.exp.exp_cuda_seed import set_seed  # reproducibility
from domainlab.exp.exp_main import Exp
from domainlab.exp_protocol import aggregate_results
from domainlab.utils.generate_benchmark_plots import gen_benchmark_plots


def domainlab_cli():
    """
    Function used to run domainlab as a command line tool for the package installed with pip.
    """
    args = parse_cmd_args()
    if args.bm_dir:
        aggregate_results.agg_main(args.bm_dir)
    elif args.plot_data is not None:
        gen_benchmark_plots(
            args.plot_data, args.outp_dir, use_param_index=args.param_idx
        )
    else:
        set_seed(args.seed)
        exp = Exp(args=args)
        exp.execute()



"""
Upon pixel wise mean and variance
"""
import torch

from domainlab import g_inst_component_loss_agg


class NLLPixelLogistic256(object):
    """
    Compute pixel wise negative likelihood of image,
    given pixel wise mean and variance.
    Pixel intensity is divided into bins of 256 levels.
    p.d.f. is calculated through
    c.d.f.(x_{i,j}+bin_size/scale) - c.d.f.(x_{i,j})
    # https://github.com/openai/iaf/blob/master/tf_utils/distributions.py#L29
    """

    def __init__(self, reduce_dims=(1, 2, 3), bin_size=1.0 / 256.0):
        """
        :param reduce_dims:
        """
        self.reduce_dims = reduce_dims
        self.bin_size = bin_size

    def __call__(self, tensor, mean, logvar):
        """
        :math: `p(tensor) = prod_{i,j}p(tensor_{i,j}, mu_{i,j}, sigma_{i,j})`
        The reconstruction likelihood should always be summation of all pixels
        :param tensor:
        :param mean:
        :param logvar:
        """
        scale = torch.exp(logvar)
        tensor = (torch.floor(tensor / self.bin_size) * self.bin_size - mean) / scale
        cdf_plus = torch.sigmoid(tensor + self.bin_size / scale)
        cdf_minus = torch.sigmoid(tensor)

        # negative log-likelihood for each pixel
        log_logist_256 = -torch.log(cdf_plus - cdf_minus + 1.0e-7)
        # torch.Size([100, 3, 28, 28])
        nll = g_inst_component_loss_agg(log_logist_256, dim=self.reduce_dims)
        # NOTE: pixel NLL should always be summed
        # across the whole image of all channels
        # NOTE: result should be order 1 tensor of dim batch_size
        return nll



"""
update hyper-parameters during training
"""
from domainlab.algos.trainers.hyper_scheduler import HyperSchedulerWarmupLinear
from domainlab.algos.trainers.train_basic import TrainerBasic
from domainlab.utils.logger import Logger


class TrainerHyperScheduler(TrainerBasic):
    """
    TrainerHyperScheduler
    """

    def set_scheduler(
        self, scheduler, total_steps, flag_update_epoch=False, flag_update_batch=False
    ):
        """
        set the warmup strategy from objective scheduler
        set wheter the hyper-parameter scheduling happens per epoch or per batch

        Args:
            scheduler: The class name of the scheduler, the object corresponding to
            this class name will be created inside model
            total_steps: number of steps to change the hyper-parameters
            flag_update_epoch: if hyper-parameters should be changed per epoch
            flag_update_batch: if hyper-parameters should be changed per batch
        """
        self.hyper_scheduler = self.model.hyper_init(scheduler)
        # let model register its hyper-parameters to the scheduler
        self.flag_update_hyper_per_epoch = flag_update_epoch
        self.flag_update_hyper_per_batch = flag_update_batch
        self.hyper_scheduler.set_steps(total_steps=total_steps)

    def before_batch(self, epoch, ind_batch):
        """
        if hyper-parameters should be updated per batch, then step
        should be set to epoch*self.num_batches + ind_batch
        """
        if self.flag_update_hyper_per_batch:
            self.model.hyper_update(
                epoch * self.num_batches + ind_batch, self.hyper_scheduler
            )
        return super().before_batch(epoch, ind_batch)

    def before_tr(self):
        if self.hyper_scheduler is None:
            logger = Logger.get_logger()
            logger.warning(
                "hyper-parameter scheduler not set,"
                "going to use default Warmpup and epoch update"
            )
            self.set_scheduler(
                HyperSchedulerWarmupLinear,
                total_steps=self.aconf.warmup,
                flag_update_epoch=True,
            )

    def tr_epoch(self, epoch):
        """
        update hyper-parameters only per epoch
        """
        if self.flag_update_hyper_per_epoch:
            self.model.hyper_update(epoch, self.hyper_scheduler)
        return super().tr_epoch(epoch)



"""
make an experiment
"""
from domainlab.arg_parser import mk_parser_main
from domainlab.exp.exp_main import Exp


def mk_exp(task, model, trainer: str, test_domain: str, batchsize: int, nocu=False):
    """
    Creates a custom experiment. The user can specify the input parameters.

    Input Parameters:
        - task: create a task to a custom dataset by importing "mk_task_dset" function from
        "domainlab.tasks.task_dset". For more explanation on the input params refer to the
        documentation found in "domainlab.tasks.task_dset.py".
        - model: create a model [NameOfModel] by importing "mk_[NameOfModel]" function from
        "domainlab.models.model_[NameOfModel]". For a concrete example and explanation of the input
        params refer to the documentation found in "domainlab.models.model_[NameOfModel].py"
        - trainer: string,
        - test_domain: string,
        - batch size: int

    Returns: experiment
    """
    str_arg = (
        f"--model=apimodel --trainer={trainer} --te_d={test_domain} --bs={batchsize}"
    )
    if nocu:
        str_arg += " --nocu"
    parser = mk_parser_main()
    conf = parser.parse_args(str_arg.split())
    exp = Exp(conf, task, model=model)
    return exp



import os

import torch.utils.data as data

from domainlab.dsets.utils_data import fun_img_path_loader_default
from domainlab.utils.utils_class import store_args


class DsetImPathList(data.Dataset):
    @store_args
    def __init__(self, root_img, path2filelist, trans_img=None, trans_target=None):
        """
        one file provide image path and label which forms a domain
        """
        self.list_tuple_img_label = []
        self.get_list_tuple_img_label()

    def get_list_tuple_img_label(self):
        with open(self.path2filelist, "r") as f_h:
            for str_line in f_h.readlines():
                path_img, label_img = str_line.strip().split()
                self.list_tuple_img_label.append(
                    (path_img, int(label_img))
                )  # @FIXME: string to int, not necessarily continuous

    def __getitem__(self, index):
        path_img, target = self.list_tuple_img_label[index]
        target = target - 1  # @FIXME:  make this more general
        img = fun_img_path_loader_default(os.path.join(self.root_img, path_img))
        if self.trans_img is not None:
            img = self.trans_img(img)
        if self.trans_target is not None:
            target = self.trans_target(target)
        return img, target, path_img

    def __len__(self):
        return len(self.list_tuple_img_label)



"""
Functions to join the csv result files generate by different jobs of the benchmarks into
a single csv file.
"""

import os
from typing import List

from domainlab.utils.generate_benchmark_plots import gen_benchmark_plots


def agg_results(input_files: List[str], output_file: str):
    """
    Aggregrates the results of the snakemake benchmark.

    Combines csv files with identical columns into a single csv file.

    :param input_files: List of csv files with identical header.
    :param output_file: Output csv file.
    """
    os.makedirs(os.path.dirname(output_file), exist_ok=True)
    has_header = False
    # logger = Logger.get_logger()
    # logger.debug(f"exp_results={input.exp_results}")
    with open(output_file, "w") as out_stream:
        for res in input_files:
            with open(res, "r") as in_stream:
                if has_header:
                    # skip header line
                    in_stream.readline()
                else:
                    out_stream.writelines(in_stream.readline())
                    has_header = True
                # write results to common file.
                out_stream.writelines(in_stream.readlines())


def agg_from_directory(input_dir: str, output_file: str):
    """Aggregates all results from a directory. Used to aggregate partial results."""
    file_list = [input_dir + os.sep + f for f in os.listdir(input_dir)]
    agg_results(file_list, output_file)


def agg_main(bm_dir: str, skip_plotting: bool = False):
    """Aggregates partial results and generate plots."""
    agg_output = f"{bm_dir}/results.csv"
    agg_input = f"{bm_dir}/rule_results"
    agg_from_directory(agg_input, agg_output)
    gen_benchmark_plots(agg_output, f"{bm_dir}/graphics", skip_plotting)



import torch
import torch.distributions as dist
import torch.nn as nn


class LSCondPriorLinearBnReluLinearSoftPlus(nn.Module):
    """
    Location-Scale: from hyper-prior to current layer prior distribution
    """

    def __init__(self, hyper_prior_dim, z_dim, hidden_dim=None):
        super().__init__()
        if hidden_dim is None:
            self.hidden_dim = z_dim
        self.net_linear_bn_relu = nn.Sequential(
            nn.Linear(hyper_prior_dim, self.hidden_dim, bias=False),
            nn.BatchNorm1d(self.hidden_dim),
            nn.ReLU(),
        )
        self.fc_loc = nn.Sequential(nn.Linear(self.hidden_dim, z_dim))
        # No activation, because latent code z variable can take both negative and positive value
        self.fc_scale = nn.Sequential(nn.Linear(self.hidden_dim, z_dim), nn.Softplus())

        # initialization
        torch.nn.init.xavier_uniform_(self.net_linear_bn_relu[0].weight)
        torch.nn.init.xavier_uniform_(self.fc_loc[0].weight)
        self.fc_loc[0].bias.data.zero_()  # No Bias
        torch.nn.init.xavier_uniform_(self.fc_scale[0].weight)
        self.fc_scale[0].bias.data.zero_()  # No Bias

    def forward(self, hyper_prior):
        """
        from hyper-prior value to current latent variable distribution
        """
        hidden = self.net_linear_bn_relu(hyper_prior)
        z_loc = self.fc_loc(hidden)
        z_scale = self.fc_scale(hidden) + 1e-7
        prior = dist.Normal(z_loc, z_scale)
        return prior
